# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""%(prog)s is a command to manipulate ROS workspaces. %(prog)s replaces its predecessor rosinstall.

Official usage:
  %(prog)s CMD [ARGS] [OPTIONS]

%(prog)s will try to infer install path from context

Type '%(prog)s help' for usage.
"""

from __future__ import print_function
import os
import sys
import yaml

from optparse import OptionParser

from wstool.cli_common import get_info_list, get_info_table, \
    get_info_table_raw_csv, get_workspace, ONLY_OPTION_VALID_ATTRS
import rosinstall.rosinstall_cmd as rosinstall_cmd
from wstool.multiproject_cmd import get_config, cmd_install_or_update, \
    cmd_snapshot, cmd_version, cmd_info, cmd_find_unmanaged_repos
import rosinstall.__version__

from wstool.common import MultiProjectException, select_elements
from wstool.helpers import ROSINSTALL_FILENAME
from rosinstall.helpers import get_ros_package_path, get_ros_stack_path
from wstool.multiproject_cli import MultiprojectCLI, __MULTIPRO_CMD_DICT__, \
    __MULTIPRO_CMD_HELP_LIST__, __MULTIPRO_CMD_ALIASES__, \
    IndentedHelpFormatterWithNL, list_usage

## This file adds or extends commands from multiproject_cli where ROS
## specific output has to be generated.

# extend the commands of multiproject
__ROSWS_CMD_DICT__ = {}
__ROSWS_CMD_DICT__.update(__MULTIPRO_CMD_DICT__)
__ROSWS_CMD_DICT__["regenerate"] = "create ROS workspace specific setup files"

__ROSWS_CMD_HELP_LIST__ = __MULTIPRO_CMD_HELP_LIST__[:]
__ROSWS_CMD_HELP_LIST__.extend([None, 'regenerate'])

_PROGNAME = 'rosws'
_VARNAME = 'ROS_WORKSPACE'


class RoswsCLI(MultiprojectCLI):

    def __init__(self, config_filename=ROSINSTALL_FILENAME, progname=_PROGNAME):
        MultiprojectCLI.__init__(self,
                                 progname=progname,
                                 config_filename=config_filename,
                                 allow_other_element=True,
                                 config_generator=rosinstall_cmd.cmd_persist_config)

    def cmd_init(self, argv):
        if self.config_filename is None:
            print('Error: Bug: config filename required for init')
            return 1
        parser = OptionParser(
            usage="""usage: %s init [TARGET_PATH [SOURCE_PATH]]?""" % self.progname,
            formatter=IndentedHelpFormatterWithNL(),
            description=__MULTIPRO_CMD_DICT__["init"] + """

%(prog)s init does the following:
  1. Reads folder/file/web-uri SOURCE_PATH looking for a rosinstall yaml
  2. Creates new %(cfg_file)s file at TARGET-PATH
  3. Generates ROS setup files

SOURCE_PATH can e.g. be a folder like /opt/ros/electric
If PATH is not given, uses current dir.

Examples:
$ %(prog)s init ~/fuerte /opt/ros/fuerte
""" % {'cfg_file': self.config_filename, 'prog': self.progname},
            epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        parser.add_option("-c", "--catkin", dest="catkin", default=False,
                          help="Declare this is a catkin build.",
                          action="store_true")
        parser.add_option("--cmake-prefix-path", dest="catkinpp", default=None,
                          help="Where to set the CMAKE_PREFIX_PATH",
                          action="store")
        parser.add_option("--continue-on-error", dest="robust", default=False,
                          help="Continue despite checkout errors",
                          action="store_true")
        parser.add_option("-j", "--parallel", dest="jobs", default=1,
                          help="How many parallel threads to use for installing",
                          action="store")
        (options, args) = parser.parse_args(argv)
        if len(args) < 1:
            target_path = '.'
        else:
            target_path = args[0]

        if not os.path.isdir(target_path):
            if not os.path.exists(target_path):
                os.mkdir(target_path)
            else:
                print('Error: Cannot create in target path %s ' % target_path)

        if os.path.exists(os.path.join(target_path, self.config_filename)):
            print('Error: There already is a workspace config file %s at "%s". Use %s install/modify.' %
                  (self.config_filename, target_path, self.progname))
            return 1
        if len(args) > 2:
            parser.error('Too many arguments')
        config_uris = []
        if len(args) == 2:
            config_uris.append(args[1])
        if len(config_uris) > 0:
            print('Using ROS_ROOT: %s' % config_uris[0])

        config = get_config(basepath=target_path,
                            additional_uris=config_uris,
                            config_filename=self.config_filename)

        # includes ROS specific files

        print("Writing %s" % os.path.join(config.get_base_path(),
              self.config_filename))
        rosinstall_cmd.cmd_persist_config(config)

        ## install or update each element
        install_success = cmd_install_or_update(
            config,
            robust=False,
            num_threads=int(options.jobs))

        rosinstall_cmd.cmd_generate_ros_files(config,
                                              target_path,
                                              nobuild=True,
                                              rosdep_yes=False,
                                              catkin=options.catkin,
                                              catkinpp=options.catkinpp,
                                              no_ros_allowed=True)

        if not install_success:
            print("Warning: installation encountered errors, but --continue-on-error was requested.  Look above for warnings.")
        print("\nrosinstall update complete.")
        if (options.catkin is False
            and options.catkinpp is None):
            print("\nType 'source %s/setup.bash' to change into this environment. Add that source command to the bottom of your ~/.bashrc to set it up every time you log in.\n\nIf you are not using bash please see http://www.ros.org/wiki/rosinstall/NonBashShells " % os.path.abspath(target_path))
        return 0

    def cmd_regenerate(self, target_path, argv, config=None):
        parser = OptionParser(usage="usage: %s regenerate" % self.progname,
                              formatter=IndentedHelpFormatterWithNL(),
                              description=__MULTIPRO_CMD_DICT__["remove"] + """

this command without options generates files setup.sh, setup.bash and
setup.zsh. Note that doing this is unnecessary in general, as these
files do not change anymore, unless you change from one ROS distro to
another (which you should never do like this, create a separate new
workspace instead), or you deleted or modified any of those files
accidentally.
""",
                              epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        parser.add_option("-c", "--catkin", dest="catkin", default=False,
                          help="Declare this is a catkin build.",
                          action="store_true")
        parser.add_option("--cmake-prefix-path", dest="catkinpp", default=None,
                          help="Where to set the CMAKE_PREFIX_PATH",
                          action="store")
        # -t option required here for help but used one layer above, see cli_common
        parser.add_option("-t", "--target-workspace", dest="workspace", default=None,
                          help="which workspace to use",
                          action="store")
        (options, args) = parser.parse_args(argv)
        if len(args) > 0:
            print("Error: Too many arguments.")
            print(parser.usage)
            return -1

        if config is None:
            config = get_config(
                target_path,
                additional_uris=[],
                config_filename=self.config_filename)
        elif config.get_base_path() != target_path:
            raise MultiProjectException(
                "Config path does not match %s %s " % (config.get_base_path(),
                                                       target_path))
        rosinstall_cmd.cmd_generate_ros_files(config,
                                              target_path,
                                              nobuild=True,
                                              rosdep_yes=False,
                                              catkin=options.catkin,
                                              catkinpp=options.catkinpp,
                                              no_ros_allowed=True)
        return 0

    def cmd_info(self, target_path, argv, reverse=True, config=None):
        # similar to multiproject_cli except shows ros-pkg-path
        # options
        parser = OptionParser(
            usage="usage: %s info [localname]* [OPTIONS]" % self.progname,
            formatter=IndentedHelpFormatterWithNL(),
            description=__MULTIPRO_CMD_DICT__["info"] + """

The Status (S) column shows
 x  for missing
 L  for uncommited (local) changes
 V  for difference in version and/or remote URI
 C  for difference in local and remote versions

The 'Version-Spec' column shows what tag, branch or revision was given
in the .rosinstall file. The 'UID' column shows the unique ID of the
current (and specified) version. The 'URI' column shows the configured
URL of the repo.

If status is V, the difference between what was specified and what is
real is shown in the respective column. For SVN entries, the url is
split up according to standard layout (trunk/tags/branches).  The
ROS_PACKAGE_PATH follows the order of the table, earlier entries
overlay later entries.

When given one localname, just show the data of one element in list form.
This also has the generic properties element which is usually empty.

The --only option accepts keywords: %(opts)s

Examples:
$ %(prog)s info -t ~/ros/fuerte
$ %(prog)s info robot_model
$ %(prog)s info --yaml
$ %(prog)s info --only=path,cur_uri,cur_revision robot_model geometry
""" % {'prog': self.progname, 'opts': ONLY_OPTION_VALID_ATTRS},
            epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        parser.add_option(
            "--root", dest="show_ws_root", default=False,
            help="Show workspace root path",
            action="store_true")
        parser.add_option(
            "--data-only", dest="data_only", default=False,
            help="Does not provide explanations",
            action="store_true")
        parser.add_option(
            "--no-pkg-path", dest="no_pkg_path", default=False,
            help="Suppress ROS_PACKAGE_PATH.",
            action="store_true")
        parser.add_option(
            "--pkg-path-only", dest="pkg_path_only", default=False,
            help="Shows only ROS_PACKAGE_PATH separated by ':'. Supercedes all other options.",
            action="store_true")
        parser.add_option(
            "--only", dest="only", default=False,
            help="Shows comma-separated lists of only given comma-separated attribute(s).",
            action="store")
        parser.add_option(
            "--yaml", dest="yaml", default=False,
            help="Shows only version of single entry. Intended for scripting.",
            action="store_true")
        parser.add_option(
            "--fetch", dest="fetch", default=False,
            help="When used, retrieves version information from remote (takes longer).",
            action="store_true")
        parser.add_option(
            "-u", "--untracked", dest="untracked",
            default=False,
            help="Also show untracked files as modifications",
            action="store_true")
        # -t option required here for help but used one layer above, see cli_common
        parser.add_option(
            "-t", "--target-workspace", dest="workspace", default=None,
            help="which workspace to use",
            action="store")
        parser.add_option(
            "-m", "--managed-only", dest="unmanaged", default=True,
            help="only show managed elements",
            action="store_false")
        (options, args) = parser.parse_args(argv)

        if config is None:
            config = get_config(
                target_path,
                additional_uris=[],
                config_filename=self.config_filename)
        elif config.get_base_path() != target_path:
            raise MultiProjectException("Config path does not match %s %s " %
                                        (config.get_base_path(), target_path))

        if options.show_ws_root:
            print(config.get_base_path())
            return 0

        if args == []:
            args = None

        if options.pkg_path_only:
            print(":".join(get_ros_package_path(config)))
            return 0
        if options.no_pkg_path:
            header = 'workspace: %s\nROS_ROOT: %s' % (target_path,
                                                      get_ros_stack_path(config))
            print(header)
            return 0
        elif options.only:
            only_options = options.only.split(",")
            if only_options == '':
                parser.error('No valid options given')
            lines = get_info_table_raw_csv(config,
                                           properties=only_options,
                                           localnames=args)
            print('\n'.join(lines))
            return 0
        elif options.yaml:
            source_aggregate = cmd_snapshot(config, localnames=args)
            print(yaml.safe_dump(source_aggregate), end='')
            return 0

        # this call takes long, as it invokes scms.
        outputs = cmd_info(config, localnames=args,
                           untracked=options.untracked,
                           fetch=options.fetch)
        if args and len(args) == 1:
            # if only one element selected, print just one line
            print(get_info_list(config.get_base_path(),
                                outputs[0],
                                options.data_only))
            return 0

        header = 'workspace: %s\nROS_ROOT: %s' % (target_path,
                                                  get_ros_stack_path(config))
        print(header)
        table = get_info_table(config.get_base_path(),
                               outputs,
                               options.data_only,
                               reverse=reverse)
        if table is not None and table != '':
            print("\n%s" % table)

        if options.unmanaged:
            outputs2 = cmd_find_unmanaged_repos(config)
            table2 = get_info_table(config.get_base_path(),
                                   outputs2,
                                   options.data_only,
                                   reverse=reverse,
                                   unmanaged=True)
            if table2 is not None and table2 != '':
                print("\nAlso detected these repositories in the workspace, add using '%s set':\n\n%s" % (self.progname, table2))

        return 0


def rosws_main(argv=None, usage=None):
    """
    Calls the function corresponding to the first argument.

    :param argv: sys.argv by default
    :param usage: function printing usage string, multiproject_cli.list_usage by default
    """
    if argv is None:
        argv = sys.argv
    if (sys.argv[0] == '-c'):
        sys.argv = [_PROGNAME] + sys.argv[1:]
    if '--version' in argv:
        print("%s: \t%s\n%s" %
              (_PROGNAME, rosinstall.__version__.version, cmd_version()))
        sys.exit(0)

    if not usage:
        usage = lambda: print(list_usage(progname=_PROGNAME,
                                         description=__doc__,
                                         command_keys=__ROSWS_CMD_HELP_LIST__,
                                         command_helps=__ROSWS_CMD_DICT__,
                                         command_aliases=__MULTIPRO_CMD_ALIASES__))
    workspace = None
    if len(argv) < 2:
        try:
            workspace = get_workspace(argv,
                                      os.getcwd(),
                                      config_filename=ROSINSTALL_FILENAME,
                                      varname=_VARNAME)
            argv.append('info')
        except MultiProjectException as e:
            print(str(e))
            usage()
            return 0

    if argv[1] in ['--help', '-h']:
        usage()
        return 0

    try:
        command = argv[1]
        args = argv[2:]

        if command == 'help':
            if len(argv) < 3:
                usage()
                return 0

            else:
                command = argv[2]
                args = argv[3:]
                args.insert(0, "--help")
                # help help
                if command == 'help':
                    usage()
                    return 0
        cli = RoswsCLI()

        # commands for which we do not infer target workspace
        commands = {'init': cli.cmd_init}
        # commands which work on a workspace
        ws_commands = {
            'info': cli.cmd_info,
            'remove': cli.cmd_remove,
            'regenerate': cli.cmd_regenerate,
            'set': cli.cmd_set,
            'merge': cli.cmd_merge,
            'foreach': cli.cmd_foreach,
            'scrape': cli.cmd_scrape,
            'diff': cli.cmd_diff,
            'status': cli.cmd_status,
            'update': cli.cmd_update}
        for label in list(ws_commands.keys()):
            alias = __MULTIPRO_CMD_ALIASES__.get(label, None)
            if alias:
                ws_commands[alias] = ws_commands[label]

        if command not in commands and command not in ws_commands:
            if os.path.exists(command):
                args = ['-t', command] + args
                command = 'info'
            else:
                if command.startswith('-'):
                    print("First argument must be name of a command: %s" %
                          command)
                else:
                    print("Error: unknown command: %s" % command)
                usage()
                return 1

        if command in commands:
            return commands[command](args)
        else:
            if workspace is None and not '--help' in args and not '-h' in args:
                workspace = get_workspace(args,
                                          os.getcwd(),
                                          config_filename=ROSINSTALL_FILENAME,
                                          varname=_VARNAME)
            return ws_commands[command](workspace, args)

    except KeyboardInterrupt:
        return 1



# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import rosinstall.__version__

from rosinstall.helpers import ROSInstallException, get_ros_stack_path

# template for catkin fuerte, not valid for Groovy and beyond, to be
# removed once fuerte goes out of support
CATKIN_CMAKE_TOPLEVEL = """#
#  TOPLEVEL cmakelists
#
cmake_minimum_required(VERSION 2.8)
cmake_policy(SET CMP0003 NEW)
cmake_policy(SET CMP0011 NEW)

set(CMAKE_CXX_FLAGS_INIT "-Wall")

enable_testing()

include(${CMAKE_SOURCE_DIR}/workspace-config.cmake OPTIONAL)

list(APPEND CMAKE_PREFIX_PATH ${CMAKE_BINARY_DIR} ${CMAKE_BINARY_DIR}/cmake)

file(MAKE_DIRECTORY ${CMAKE_BINARY_DIR}/lib)
file(MAKE_DIRECTORY ${CMAKE_BINARY_DIR}/bin)

if (IS_DIRECTORY ${CMAKE_SOURCE_DIR}/catkin)
  message(STATUS "+++ catkin")
  set(CATKIN_BUILD_PROJECTS "ALL" CACHE STRING
    "List of projects to build, or ALL for all.  Use to completely exclude certain projects from cmake traversal.")
  add_subdirectory(catkin)
else()
  find_package(catkin)
endif()

catkin_workspace()
"""

SHELL_HEADER = """# THIS IS AN AUTO-GENERATED FILE
# IT IS UNLIKELY YOU WANT TO EDIT THIS FILE BY HAND
# IF YOU WANT TO CHANGE THE ROS ENVIRONMENT VARIABLES
# USE THE rosinstall OR rosws TOOL INSTEAD.
# Generator version: %s
# see: http://www.ros.org/wiki/rosinstall
""" % rosinstall.__version__.version


def generate_catkin_cmake(path, catkinpp):
    with open(os.path.join(path, "CMakeLists.txt"), 'w') as cmake_file:
        cmake_file.write(CATKIN_CMAKE_TOPLEVEL)

    if catkinpp:
        with open(os.path.join(path, "workspace-config.cmake"), 'w') as config_file:
            config_file.write("set (CMAKE_PREFIX_PATH %s)" % catkinpp)


def generate_embedded_python():
    return """import sys
import os
import yaml

workspace_path = os.environ.get('ROS_WORKSPACE', os.path.abspath('.'))
filename = os.path.join(workspace_path, '.rosinstall')

if not os.path.isfile(filename):
  print('ERROR')
  sys.exit("There is no file at %s" % filename)

with open(filename, "r") as fhand:
  try:
    v = fhand.read();
  except Exception as e:
    print('ERROR')
    sys.exit("Failed to read file: %s %s " % (filename, str(e)))

try:
  y = yaml.load(v);
except Exception as e:
  print('ERROR')
  sys.exit("Invalid yaml in %s: %s " % (filename, str(e)))

if y is not None:

  # put all non-setupfile entries into ROS_PACKAGE_PATH
  paths = []
  for vdict in y:
    for k, v in vdict.items():
      if v is not None and k != "setup-file":
        path = os.path.join(workspace_path, v['local-name'])
        if not os.path.isfile(path):
          # add absolute path from workspace to relative paths
          paths.append(os.path.normpath(path))
        else:
          print('ERROR')
          sys.exit("ERROR: referenced path is a file, not a folder: %s" % path)
  output = ''
  # add paths in reverse order
  if len(paths) > 0:
    output += ':'.join(reversed(paths))

  # We also want to return the location of any setupfile elements
  output += 'ROSINSTALL_PATH_SETUPFILE_SEPARATOR'
  setupfile_paths = []
  for vdict in y:
    for k, v in vdict.items():
      if v is not None and k == "setup-file":
        path = os.path.join(workspace_path, v['local-name'])
        if not os.path.exists(path):
          print('ERROR')
          sys.exit("WARNING: referenced setupfile does not exist: %s" % path)
        elif os.path.isfile(path):
          setupfile_paths.append(path)
        else:
          print('ERROR')
          sys.exit("ERROR: referenced setupfile is a folder: %s" % path)
  output += ':'.join(setupfile_paths)

  # printing will store the result in the variable
  print(output)"""


def generate_setup_sh_text(workspacepath):
    '''
    generates the string that goes into setup.sh.

    Sadly we cannot infer the workspacepath from within the sourced
    file, previous hacks trying to determine it from the shell context
    all failed in corner cases.

    :param workspacepath: The path to the workspace
    '''

    pycode = generate_embedded_python()

    # overlay or standard
    text = """#!/usr/bin/env sh
%(header)s

# This setup.sh file has to parse .rosinstall file, and source similar
# setup.sh files recursively. In the course of recursion, shell
# variables get overwritten. This means that when returning from
# recursion, any variable may be in a different state

# These variables accumulate data through recursion and must only be
# reset and unset at the top level of recursion.

if [ x"$_ROSINSTALL_IN_RECURSION" != x"recurse" ] ; then
  # reset setupfile accumulator
  _SETUPFILES_ROSINSTALL=
  _ROS_PACKAGE_PATH_ROSINSTALL=
  # reset RPP before sourcing other setup files
  export ROS_PACKAGE_PATH=
fi

export ROS_WORKSPACE=%(wspath)s
if [ ! "$ROS_MASTER_URI" ] ; then export ROS_MASTER_URI=http://localhost:11311 ; fi
unset ROS_ROOT

unset _SETUP_SH_ERROR

# python script to read .rosinstall even when rosinstall is not installed
# this files parses the .rosinstall and sets environment variables accordingly
# The ROS_PACKAGE_PATH contains all elements in reversed order (for historic reasons)

# We store into _PARSED_CONFIG the result of python code,
# which is the ros_package_path and the list of setup_files to source
# Using python here to benefit of the pyyaml library
export _PARSED_CONFIG=`/usr/bin/env python << EOPYTHON

%(pycode)s
EOPYTHON`

if [ x"$_PARSED_CONFIG" = x"ERROR" ]; then
  echo 'Could not parse .rosinstall file' 1<&2
  _SETUP_SH_ERROR=1
fi

# using sed to split up ros_package_path and setupfile results
_ROS_PACKAGE_PATH_ROSINSTALL_NEW=`echo "$_PARSED_CONFIG" | sed 's,\(.*\)ROSINSTALL_PATH_SETUPFILE_SEPARATOR\(.*\),\\1,'`
if [ ! -z "$_ROS_PACKAGE_PATH_ROSINSTALL_NEW" ]; then
  if [ ! -z "$_ROS_PACKAGE_PATH_ROSINSTALL" ]; then
    export _ROS_PACKAGE_PATH_ROSINSTALL=$_ROS_PACKAGE_PATH_ROSINSTALL:$_ROS_PACKAGE_PATH_ROSINSTALL_NEW
  else
    export _ROS_PACKAGE_PATH_ROSINSTALL=$_ROS_PACKAGE_PATH_ROSINSTALL_NEW
  fi
fi

_SETUPFILES_ROSINSTALL_NEW=`echo "$_PARSED_CONFIG" | sed 's,\(.*\)'ROSINSTALL_PATH_SETUPFILE_SEPARATOR'\(.*\),\\2,'`
if [ ! -z "$_SETUPFILES_ROSINSTALL_NEW" ]; then
  if [ ! -z "$_SETUPFILES_ROSINSTALL" ]; then
    _SETUPFILES_ROSINSTALL=$_SETUPFILES_ROSINSTALL_NEW:$_SETUPFILES_ROSINSTALL
  else
    _SETUPFILES_ROSINSTALL=$_SETUPFILES_ROSINSTALL_NEW
  fi
fi
unset _PARSED_CONFIG

# colon separates entries
_LOOP_SETUP_FILE=`echo $_SETUPFILES_ROSINSTALL | sed 's,\([^:]*\)[:]\(.*\),\\1,'`
# this loop does fake recursion, as the called setup.sh may work on
# the remaining elements in the _SETUPFILES_ROSINSTALL stack
while [ ! -z "$_LOOP_SETUP_FILE" ]
do
  # need to pop from stack before recursing, as chained setup.sh might rely on this
  _SETUPFILES_ROSINSTALL=`echo $_SETUPFILES_ROSINSTALL | sed 's,\([^:]*[:]*\),,'`
  if [ -f "$_LOOP_SETUP_FILE" ]; then
    _ROSINSTALL_IN_RECURSION=recurse
    . $_LOOP_SETUP_FILE
    unset _ROSINSTALL_IN_RECURSION
  else
    echo warn: no such file : "$_LOOP_SETUP_FILE"
  fi
  _LOOP_SETUP_FILE=`echo $_SETUPFILES_ROSINSTALL | sed 's,\([^:]*\)[:]\(.*\),\\1,'`
done

unset _LOOP_SETUP_FILE
unset _SETUPFILES_ROSINSTALL

# prepend elements from .rosinstall file to ROS_PACKAGE_PATH
# ignoring duplicates entries from value set by setup files
export ROS_PACKAGE_PATH=`/usr/bin/env python << EOPYTHON
import os
ros_package_path = os.environ.get('ROS_PACKAGE_PATH', '')
original_elements = ros_package_path.split(':')
ros_package_path2 = os.environ.get('_ROS_PACKAGE_PATH_ROSINSTALL', '')
new_elements = ros_package_path2.split(':')
new_elements = [path for path in new_elements if path]

for original_path in original_elements:
  if original_path and original_path not in new_elements:
    new_elements.append(original_path)
print(':'.join(new_elements))
EOPYTHON`

unset _ROS_PACKAGE_PATH_ROSINSTALL

# restore ROS_WORKSPACE in case other setup.sh changed/unset it
export ROS_WORKSPACE=%(wspath)s

# if setup.sh did not set ROS_ROOT (pre-fuerte)
if [ -z "${ROS_ROOT}" ]; then
  # using ROS_ROOT now being in ROS_PACKAGE_PATH
  export _ROS_ROOT_ROSINSTALL=`/usr/bin/env python << EOPYTHON
import sys, os;
if 'ROS_PACKAGE_PATH' in os.environ:
  pkg_path = os.environ['ROS_PACKAGE_PATH']
  for path in pkg_path.split(':'):
    if (os.path.basename(path) == 'ros'
        and os.path.isfile(os.path.join(path, 'stack.xml'))):
      print(path)
      break
EOPYTHON`

  if [ ! -z "${_ROS_ROOT_ROSINSTALL}" ]; then
    export ROS_ROOT=$_ROS_ROOT_ROSINSTALL
    export PATH=$ROS_ROOT/bin:$PATH
    export PYTHONPATH=$ROS_ROOT/core/roslib/src:$PYTHONPATH
  fi
unset _ROS_ROOT_ROSINSTALL
fi

if [ ! -z "$_SETUP_SH_ERROR" ]; then
  # return failure code when sourcing file
  false
fi
""" % {'header': SHELL_HEADER, 'wspath': workspacepath, 'pycode': pycode}

    return text


def generate_setup_bash_text(shell):
    '''
    Generates the contents that go into a setup.bash or setup.zsh
    file.  The intent of such a file is to enable shell extensions,
    such as special ros commands and tab completion.  The generation
    is complex because the setup of the system changed between ROS
    electric and fuerte. In fuerte, the distro setup.sh also loads
    distro rosbash based on CATKIN_SHELL. Before fuerte, it is up to
    setup.bash to do so.
    '''
    if shell == 'bash':
        script_path = """
SCRIPT_PATH="${BASH_SOURCE[0]}";
if([ -h "${SCRIPT_PATH}" ]) then
  while([ -h "${SCRIPT_PATH}" ]) do SCRIPT_PATH=`readlink "${SCRIPT_PATH}"`; done
fi
export OLDPWDBAK=$OLDPWD
pushd . > /dev/null
cd `dirname ${SCRIPT_PATH}` > /dev/null
SCRIPT_PATH=`pwd`;
popd  > /dev/null
export OLDPWD=$OLDPWDBAK
"""
        call_setup_sh = ". $SCRIPT_PATH/setup.sh"
    elif shell == 'zsh':
        script_path = 'SCRIPT_PATH="$(dirname $0)"'
        call_setup_sh = """
emulate sh # emulate POSIX
. $SCRIPT_PATH/setup.sh
emulate zsh # back in zsh
"""
    else:
        raise ROSInstallException("%s shell unsupported." % shell)

    text = """#!/usr/bin/env %(shell)s
%(header)s

CATKIN_SHELL=%(shell)s

%(script_path)s

# Load the path of this particular setup.%(shell)s

if [ ! -f "$SCRIPT_PATH/setup.sh" ]; then
  echo "Bug: shell script unable to determine its own location: $SCRIPT_PATH"
  return 22
fi

# unset _ros_decode_path (function of rosbash) to check later whether setup.sh has sourced ros%(shell)s
unset -f _ros_decode_path 1> /dev/null 2>&1

%(call_setup_sh)s

# if we have a ROS_ROOT, then we might need to source rosbash (pre-fuerte)
if [ ! -z "${ROS_ROOT}" ]; then
  # check whether setup.sh also already sourced rosbash
  # Cannot rely on $? due to set -o errexit in build scripts
  RETURNCODE=`type _ros_decode_path 2> /dev/null | grep function 1>/dev/null 2>&1 || echo error`

  # for ROS electric and before, source rosbash
  if [ ! "$RETURNCODE" = "" ]; then
    RETURNCODE=`rospack help 1> /dev/null 2>&1 || echo error`
    if  [ "$RETURNCODE" = "" ]; then
      ROSSHELL_PATH=`rospack find rosbash`/ros%(shell)s
      if [ -e "$ROSSHELL_PATH" ]; then
        . $ROSSHELL_PATH
      fi
    else
      echo "rospack could not be found, you cannot have ros%(shell)s features until you bootstrap ros"
    fi
  fi
fi
""" % {'shell': shell,
       'script_path': script_path,
       'call_setup_sh': call_setup_sh,
       'header': SHELL_HEADER}
    return text


def generate_setup(config, no_ros_allowed=False):
    ros_root = get_ros_stack_path(config)
    if ros_root is None:
        if not no_ros_allowed:
            candidates = []
            for t in config.get_config_elements():
                if os.path.basename(t.get_local_name()) == 'ros':
                    candidates.append(t.get_path())
            raise ROSInstallException("""
No 'ros' stack detected in candidates %s.
Please add the location of a ros distribution to this command.

See http://ros.org/wiki/rosinstall.""" % (candidates))

    text = generate_setup_sh_text(workspacepath=config.get_base_path())
    setup_path = os.path.join(config.get_base_path(), 'setup.sh')
    with open(setup_path, 'w') as fhand:
        fhand.write(text)

    for shell in ['bash', 'zsh']:
        text = generate_setup_bash_text(shell)
        setup_path = os.path.join(config.get_base_path(), 'setup.%s' % shell)
        with open(setup_path, 'w') as fhand:
            fhand.write(text)



# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import subprocess
import wstool
import wstool.helpers
from test.io_wrapper import StringIO
import wstool.multiproject_cmd

from test.scm_test_base import AbstractFakeRosBasedTest
from rosinstall.rosws_cli import RoswsCLI


class RosWsTest(AbstractFakeRosBasedTest):

    def setUp(self):
        """runs wstool with generated self.simple_rosinstall to create local wstool env
        and creates a directory for a second local wstool env"""
        AbstractFakeRosBasedTest.setUp(self)

    def test_init(self):
        workspace = os.path.join(self.test_root_path, 'ws1')
        cli = RoswsCLI()
        try:
            cli.cmd_init([workspace, 'foo', 'bar'])
            fail("expected exit")
        except SystemExit:
            pass
        self.assertEqual(0, cli.cmd_init([workspace, self.simple_rosinstall]))
        self.assertTrue(os.path.exists(workspace))
        self.assertTrue(os.path.exists(os.path.join(workspace, '.rosinstall')))
        self.assertTrue(os.path.exists(os.path.join(workspace, 'setup.sh')))
        self.assertTrue(os.path.isdir(os.path.join(workspace, 'ros')))
        self.assertTrue(os.path.isdir(os.path.join(workspace, 'gitrepo')))
        self.assertTrue(os.path.isdir(os.path.join(workspace, 'gitrepo', '.git')))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(2, len(config.get_config_elements()))

    def test_init_pwd(self):
        workspace = os.path.join(self.test_root_path, 'ws1b')
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([workspace]))
        self.assertTrue(os.path.exists(workspace))
        self.assertTrue(os.path.exists(os.path.join(workspace, '.rosinstall')))
        self.assertTrue(os.path.exists(os.path.join(workspace, 'setup.sh')))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(0, len(config.get_config_elements()))

    def test_init_parallel(self):
        workspace = os.path.join(self.test_root_path, 'ws1d')
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([workspace, self.simple_rosinstall, "--parallel=5"]))
        self.assertTrue(os.path.exists(workspace))
        self.assertTrue(os.path.exists(os.path.join(workspace, '.rosinstall')))

    def test_merge(self):
        workspace = os.path.join(self.test_root_path, 'ws2')
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([workspace, self.simple_rosinstall]))
        self.assertEqual(0, cli.cmd_merge(workspace, [self.simple_changed_vcs_rosinstall, '-y']))
        self.assertFalse(os.path.isdir(os.path.join(workspace, 'hgrepo')))
        self.assertFalse(os.path.isdir(os.path.join(workspace, 'hgrepo', '.hg')))
        self.assertEqual(0, cli.cmd_update(workspace, []))
        self.assertTrue(os.path.isdir(os.path.join(workspace, 'hgrepo')))
        self.assertTrue(os.path.isdir(os.path.join(workspace, 'hgrepo', '.hg')))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(3, len(config.get_config_elements()))

    def test_remove(self):
        workspace = os.path.join(self.test_root_path, 'ws3')
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([workspace, self.simple_rosinstall]))
        self.assertEqual(0, cli.cmd_merge(workspace, [self.simple_changed_vcs_rosinstall, '-y']))
        self.assertEqual(0, cli.cmd_update(workspace, []))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(3, len(config.get_config_elements()))
        self.assertEqual(0, cli.cmd_remove(workspace, ['hgrepo']))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(2, len(config.get_config_elements()))

    def test_set_detached(self):
        workspace = os.path.join(self.test_root_path, 'ws4')
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([workspace, self.simple_rosinstall]))
        self.assertEqual(0, cli.cmd_merge(workspace, [self.simple_changed_vcs_rosinstall, '-y']))
        self.assertEqual(0, cli.cmd_update(workspace, []))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(3, len(config.get_config_elements()))
        self.assertEqual('ros', config.get_config_elements()[0].get_local_name())
        self.assertEqual('gitrepo', config.get_config_elements()[1].get_local_name())
        self.assertEqual(True, config.get_config_elements()[1].is_vcs_element())
        self.assertEqual('hgrepo', config.get_config_elements()[2].get_local_name())
        self.assertEqual(0, cli.cmd_set(workspace, [os.path.join(workspace, 'gitrepo'), '--detached', '-y']))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(3, len(config.get_config_elements()))
        self.assertEqual('ros', config.get_config_elements()[0].get_local_name())
        self.assertEqual('gitrepo', config.get_config_elements()[1].get_local_name())
        self.assertEqual(False, config.get_config_elements()[1].is_vcs_element())
        self.assertEqual('hgrepo', config.get_config_elements()[2].get_local_name())

    def test_set_add_plain(self):
        workspace = os.path.join(self.test_root_path, 'ws5')
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([workspace, self.simple_rosinstall]))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(2, len(config.get_config_elements()))
        self.assertEqual('ros', config.get_config_elements()[0].get_local_name())
        self.assertEqual('gitrepo', config.get_config_elements()[1].get_local_name())
        # detached
        self.assertEqual(0, cli.cmd_set(workspace, [os.path.join(workspace, 'foo'), '-y']))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(3, len(config.get_config_elements()))
        self.assertEqual('ros', config.get_config_elements()[0].get_local_name())
        self.assertEqual('gitrepo', config.get_config_elements()[1].get_local_name())
        self.assertEqual(False, config.get_config_elements()[2].is_vcs_element())
        self.assertEqual('foo', config.get_config_elements()[2].get_local_name())
        self.assertEqual(0, cli.cmd_set(workspace, [os.path.join(workspace, 'hgrepo'), '-y', '--detached']))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(4, len(config.get_config_elements()))
        self.assertEqual('ros', config.get_config_elements()[0].get_local_name())
        self.assertEqual('gitrepo', config.get_config_elements()[1].get_local_name())
        self.assertEqual(False, config.get_config_elements()[2].is_vcs_element())
        self.assertEqual('foo', config.get_config_elements()[2].get_local_name())
        self.assertEqual(False, config.get_config_elements()[3].is_vcs_element())
        self.assertEqual('hgrepo', config.get_config_elements()[3].get_local_name())
        # turn into scm repo
        self.assertEqual(0, cli.cmd_set(workspace, [os.path.join(workspace, 'hgrepo'), '../hgrepo', '-y', '--hg']))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(4, len(config.get_config_elements()))
        self.assertEqual('ros', config.get_config_elements()[0].get_local_name())
        self.assertEqual('gitrepo', config.get_config_elements()[1].get_local_name())
        self.assertTrue(config.get_config_elements()[3].is_vcs_element())
        self.assertEqual('hgrepo', config.get_config_elements()[3].get_local_name())

    def test_set_add_scm(self):
        workspace = os.path.join(self.test_root_path, 'ws6')
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([workspace, self.simple_rosinstall]))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(2, len(config.get_config_elements()))
        self.assertEqual('ros', config.get_config_elements()[0].get_local_name())
        self.assertEqual('gitrepo', config.get_config_elements()[1].get_local_name())

        # scm repo
        self.assertEqual(0, cli.cmd_set(workspace, [os.path.join(workspace, 'hgrepo'), '../hgrepo', '-y', '--hg']))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(3, len(config.get_config_elements()))
        self.assertEqual('ros', config.get_config_elements()[0].get_local_name())
        self.assertEqual('gitrepo', config.get_config_elements()[1].get_local_name())
        self.assertTrue(config.get_config_elements()[2].is_vcs_element())
        self.assertEqual('hgrepo', config.get_config_elements()[2].get_local_name())
        self.assertFalse(os.path.exists(os.path.join(workspace, 'hgrepo')))

        self.assertEqual(0, cli.cmd_update(workspace, []))
        self.assertTrue(os.path.exists(os.path.join(workspace, 'hgrepo')))

        path_spec = config.get_config_elements()[2].get_versioned_path_spec()
        self.assertFalse(path_spec is None)
        self.assertEqual(None, path_spec.get_version())
        self.assertEqual(None, path_spec.get_revision())
        self.assertFalse(path_spec.get_current_revision() is None)

        self.assertEqual(0, cli.cmd_set(workspace, [os.path.join(workspace, 'hgrepo'), '--version-new=0', '-y']))

        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        path_spec = config.get_config_elements()[2].get_versioned_path_spec()
        self.assertEqual('0', path_spec.get_version())
        self.assertFalse(path_spec.get_revision() is None)
        self.assertFalse(path_spec.get_current_revision() is None)
        self.assertEqual(path_spec.get_revision(), path_spec.get_current_revision())

        # change in FS to version 1
        subprocess.check_call(["touch", "hgfixed2.txt"], cwd=os.path.join(workspace, 'hgrepo'))
        subprocess.check_call(["hg", "add", "hgfixed2.txt"], cwd=os.path.join(workspace, 'hgrepo'))
        subprocess.check_call(["hg", "commit", "-m", "2nd"], cwd=os.path.join(workspace, 'hgrepo'))
        self.assertTrue(os.path.exists(os.path.join(workspace, 'hgrepo', 'hgfixed2.txt')))

        path_spec = config.get_config_elements()[2].get_versioned_path_spec()
        self.assertEqual('0', path_spec.get_version())
        self.assertFalse(path_spec.get_revision() is None)
        self.assertFalse(path_spec.get_current_revision() is None)
        self.assertNotEqual(path_spec.get_revision(), path_spec.get_current_revision())

        # revert FS to spec
        self.assertEqual(0, cli.cmd_update(workspace, []))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        path_spec = config.get_config_elements()[2].get_versioned_path_spec()
        self.assertEqual('0', path_spec.get_version())
        self.assertFalse(path_spec.get_revision() is None)
        self.assertFalse(path_spec.get_current_revision() is None)
        self.assertEqual(path_spec.get_revision(), path_spec.get_current_revision())

        # change spec to 1
        self.assertEqual(0, cli.cmd_set(workspace, [os.path.join(workspace, 'hgrepo'), '--version-new=1', '-y']))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        path_spec = config.get_config_elements()[2].get_versioned_path_spec()
        self.assertEqual('1', path_spec.get_version())
        self.assertFalse(path_spec.get_revision() is None)
        self.assertFalse(path_spec.get_current_revision() is None)
        self.assertNotEqual(path_spec.get_revision(), path_spec.get_current_revision())

        # setting version to ''
        self.assertEqual(0, cli.cmd_set(workspace, [os.path.join(workspace, 'hgrepo'), "--version-new=''", '-y']))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        path_spec = config.get_config_elements()[2].get_versioned_path_spec()
        self.assertEqual(None, path_spec.get_version())
        self.assertTrue(path_spec.get_revision() is None)
        self.assertFalse(path_spec.get_current_revision() is None)

        self.assertEqual(0, cli.cmd_update(workspace, []))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        path_spec = config.get_config_elements()[2].get_versioned_path_spec()
        self.assertEqual(None, path_spec.get_version())
        self.assertTrue(path_spec.get_revision() is None)
        self.assertFalse(path_spec.get_current_revision() is None)

    def test_info_only(self):
        workspace = os.path.join(self.test_root_path, 'ws7')
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([workspace, self.simple_rosinstall]))
        # pkg_path
        sys.stdout = output = StringIO()
        self.assertEqual(0, cli.cmd_info(workspace, ['--pkg-path-only']))
        output = output.getvalue()
        self.assertEqual(os.path.join(workspace, 'gitrepo'), output.strip())

        sys.stdout = output = StringIO()
        self.assertEqual(0, cli.cmd_info(workspace, ['--only=localname']))
        output = output.getvalue()
        self.assertEqual('ros\ngitrepo', output.strip())

        sys.stdout = output = StringIO()
        self.assertEqual(0, cli.cmd_info(workspace, ['--only=version']))
        output = output.getvalue()
        self.assertEqual('', output.strip())

        sys.stdout = output = StringIO()
        self.assertEqual(0, cli.cmd_info(workspace, ['--only=uri']))
        output = output.getvalue()
        self.assertEqual('%s\n%s\n' % (os.path.join(self.test_root_path, 'ros'), os.path.join(self.test_root_path, 'gitrepo')), output)

        sys.stdout = output = StringIO()
        self.assertEqual(0, cli.cmd_info(workspace, ['--only=cur_revision']))
        output = output.getvalue()
        self.assertEqual(82, len(output))
        sys.stdout = sys.__stdout__

        # pairs
        sys.stdout = output = StringIO();
        self.assertEqual(0, cli.cmd_info(workspace, ['--only=localname,scmtype']))
        output = output.getvalue()
        self.assertEqual('ros,git\ngitrepo,git', output.strip())
        sys.stdout = output = StringIO();
        self.assertEqual(0, cli.cmd_info(workspace, ['--only=scmtype,localname']))
        output = output.getvalue()
        self.assertEqual('git,ros\ngit,gitrepo', output.strip())

    def test_set_add_scm_change_localname(self):
        workspace = os.path.join(self.test_root_path, 'ws8')
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([workspace, self.ros_path]))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(1, len(config.get_config_elements()))
        self.assertEqual(os.path.join(self.test_root_path, 'ros'), config.get_config_elements()[0].get_local_name())

        # use a weird absolute localname
        self.assertEqual(0, cli.cmd_set(workspace, [os.path.join(workspace, '..', 'ws8', 'hgrepo'), '../hgrepo', '-y', '--hg']))
        config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
        self.assertEqual(2, len(config.get_config_elements()))
        self.assertEqual('hgrepo', config.get_config_elements()[1].get_local_name())

        oldcwd = os.getcwd()
        try:
            os.chdir(self.test_root_path)
            # try pointing to a relative dir that also exists elsewhere
            try:
                cli.cmd_set(workspace, ['gitrepo', '../gitrepo', '-y', '--hg'])
                self.fail("Expected SystemExit")
            except SystemExit:
                pass
            # use a weird relative localname
            self.assertEqual(0, cli.cmd_set(workspace, [os.path.join('ws8', 'gitrepo'), '../gitrepo', '-y', '--hg']))
            config = wstool.multiproject_cmd.get_config(workspace, config_filename='.rosinstall')
            self.assertEqual(3, len(config.get_config_elements()))

            self.assertEqual('hgrepo', config.get_config_elements()[1].get_local_name())
            self.assertEqual('gitrepo', config.get_config_elements()[2].get_local_name())
        finally:
            os.chdir(oldcwd)



import unittest
from mock import Mock
import rosinstall.locate as locate


class LocateTest(unittest.TestCase):

    def test_getters_invalid(self):
        data = {
            'vcs': None,
            'vcs_uri': None}
        self.assertRaises(locate.InvalidData,
                          locate.get_rosinstall, 'myname', data, None)
        data = {
            'vcs': None,
            'vcs_uri': 'https://code.ros.org/svn/ros-pkg'}
        self.assertRaises(locate.InvalidData,
                          locate.get_rosinstall, 'myname', data, None)
        data = {
            'vcs': 'svn',
            'vcs_uri': None}
        self.assertRaises(locate.InvalidData,
                          locate.get_rosinstall, 'myname', data, None)
        data = {}
        self.assertRaises(locate.InvalidData,
                          locate.get_rosinstall, 'myname', data, None)
        data = {'vcs_uri': 'https://code.ros.org/svn/ros-pkg'}
        self.assertRaises(locate.InvalidData,
                          locate.get_rosinstall, 'myname', data, None)
        data = {'vcs': 'svn'}
        self.assertRaises(locate.InvalidData,
                          locate.get_rosinstall, 'myname', data, None)
        data = {'rosinstalls': None}
        self.assertRaises(locate.InvalidData,
                          locate.get_rosinstall,
                          'myname', data, 'mytype', 'devel')
        data = {'rosinstall': None}
        self.assertRaises(locate.InvalidData,
                          locate.get_rosinstall,
                          'myname', data, 'mytype', None)

    def test_getters_empty(self):
        data = {'vcs': 'svn',
                'vcs_uri': 'https://code.ros.org/svn/ros-pkg'}
        self.assertEqual('', locate.get_www(None, data, None))
        self.assertEqual('', locate.get_repo(None, data, None))
        self.assertEqual('https://code.ros.org/svn/ros-pkg', locate.get_vcs_uri_for_branch(data))
        self.assertEqual('https://code.ros.org/svn/ros-pkg', locate.get_vcs_uri_for_branch(data, 'release'))
        self.assertEqual('https://code.ros.org/svn/ros-pkg', locate.get_vcs_uri_for_branch(data, 'devel'))
        self.assertEqual('svn', locate.get_vcs(None, data, None))
        self.assertEqual('- svn:\n    local-name: myname\n    uri: https://code.ros.org/svn/ros-pkg\n',
                         locate.get_rosinstall('myname',
                                               data, 'mytype'))
        self.assertEqual('- svn:\n    local-name: myname\n    uri: https://code.ros.org/svn/ros-pkg\n',
                         locate.get_rosinstall('myname',
                                               data, 'mytype', 'devel'))
        self.assertEqual('- svn:\n    local-name: myname\n    uri: https://code.ros.org/svn/ros-pkg\n',
                         locate.get_rosinstall('myname',
                                               data, 'mytype', 'release'))
        self.assertEqual('- svn:\n    local-name: foo/myname\n    uri: https://code.ros.org/svn/ros-pkg\n',
                         locate.get_rosinstall('myname',
                                               data, 'mytype', None, 'foo'))
        self.assertEqual('https://code.ros.org/svn/ros-pkg',
                         locate.get_vcs_uri(data))
        self.assertEqual('', locate.get_vcs_version(data))
        self.assertEqual('package', locate.get_type(data))

    def test_get_with_branches(self):
        data = {
            'rosinstalls':
                {'devel':
                     {'hg':
                          {'local-name': 'navigation',
                           'uri': 'https://kforge.ros.org/test/devel',
                           'version': 'navigation-1.6'}},
                 'distro':
                     {'hg':
                          {'local-name': 'navigation',
                           'uri': 'https://kforge.ros.org/test/distro',
                           'version': 'electric'}},
                 'release':
                     {'hg':
                      {'local-name': 'navigation',
                       'uri': 'https://kforge.ros.org/test/rel',
                       'version': 'navigation-1.6.5'}}},
            'vcs': 'svn',
            'vcs_uri': 'https://code.ros.org/svn/ros-pkg'}
        self.assertEqual('https://code.ros.org/svn/ros-pkg',
                         locate.get_vcs_uri_for_branch(data))
        self.assertEqual('https://kforge.ros.org/test/rel',
                         locate.get_vcs_uri_for_branch(data, 'release'))
        self.assertEqual('https://kforge.ros.org/test/devel',
                         locate.get_vcs_uri_for_branch(data, 'devel'))
        self.assertEqual('https://kforge.ros.org/test/distro',
                         locate.get_vcs_uri_for_branch(data, 'distro'))
        self.assertEqual('svn', locate.get_vcs(None, data, None))
        self.assertEqual('https://code.ros.org/svn/ros-pkg', locate.get_vcs_uri(data))
        self.assertEqual('', locate.get_vcs_version(data))
        self.assertEqual('package', locate.get_type(data))
        self.assertEqual(
            '- svn:\n    local-name: myname\n    uri: https://code.ros.org/svn/ros-pkg\n',
            locate.get_rosinstall('myname',
                                  data, 'mytype'))
        self.assertEqual(
            '- hg:\n    local-name: navigation\n    uri: https://kforge.ros.org/test/devel\n    version: navigation-1.6\n',
            locate.get_rosinstall('myname',
                                  data, 'mytype', 'devel'))
        self.assertEqual(
            '- hg:\n    local-name: navigation\n    uri: https://kforge.ros.org/test/rel\n    version: navigation-1.6.5\n',
            locate.get_rosinstall('myname',
                                  data, 'mytype', 'release'))
        self.assertEqual(
            '- svn:\n    local-name: foo/myname\n    uri: https://code.ros.org/svn/ros-pkg\n',
            locate.get_rosinstall('myname',
                                  data, 'mytype', None, 'foo'))

    def test_get_with_rosinstall(self):
        data = {
            'rosinstall':
                {'hg':
                      {'local-name': 'navigation',
                       'uri': 'https://kforge.ros.org/test/rel',
                       'version': 'navigation-1.6.5'}},
            'vcs': 'svn',
            'vcs_uri': 'https://code.ros.org/svn/ros-pkg'}
        self.assertEqual(
            '- hg:\n    local-name: navigation\n    uri: https://kforge.ros.org/test/rel\n    version: navigation-1.6.5\n',
            locate.get_rosinstall('myname',
                                  data, 'mytype'))
        self.assertEqual(
            '- hg:\n    local-name: foo/navigation\n    uri: https://kforge.ros.org/test/rel\n    version: navigation-1.6.5\n',
            locate.get_rosinstall('myname',
                                  data, 'mytype', None, 'foo'))

    def test_getters(self):
        data = {'package_type': 'package',
                'repo_name': 'visualization',
                'repo_url': '',
                'srvs': [],
                'timestamp': 1362233859.0088351,
                'url': 'http://ros.org/wiki/rviz',
                'vcs': 'svn',
                'vcs_uri': 'https://code.ros.org/svn/ros-pkg',
                'vcs_version': '0.1',
                'repository': 'navigation'}
        self.assertEqual('http://ros.org/wiki/rviz', locate.get_www(None, data, None))
        self.assertEqual('visualization', locate.get_repo(None, data, None))
        self.assertEqual(
            '- svn:\n    local-name: foo/myname\n    uri: https://code.ros.org/svn/ros-pkg\n    version: \'0.1\'\n',
            locate.get_rosinstall('myname',
                                  data, 'mytype', None, 'foo'))
        self.assertEqual('https://code.ros.org/svn/ros-pkg', locate.get_vcs_uri(data))
        self.assertEqual('0.1', locate.get_vcs_version(data))
        self.assertEqual('package', locate.get_type(data))


    def test_get_manifest_groovy(self):
        distro = 'groovy'
        # rviz
        result = locate.get_manifest('rviz', distro)
        self.assertEqual('package', result[1], result)
        # rospack
        (data, type_, url) = locate.get_manifest('rospack', distro)
        self.assertEqual('package', type_)
        data = locate._get_rosinstall_dict('rospack', data, type_)
        self.assertEqual('https://github.com/ros/rospack.git', data.get('git', {}).get('uri', ''))
        self.assertTrue(distro in data.get('git', {}).get('version', ''), data)
        # ros_comm
        (data, type_, url) = locate.get_manifest('ros_comm', distro)
        self.assertEqual('metapackage', type_)
        data = locate._get_rosinstall_dict('ros_comm', data, type_)
        self.assertEqual('https://github.com/ros/ros_comm.git', data.get('git', {}).get('uri', ''))
        self.assertTrue(distro in data.get('git', {}).get('version', ''), data)

    def test_get_manifest_current(self):
        # this test obviously returns different results over time
        distro = None
        # rviz
        result = locate.get_manifest('rviz', distro)
        self.assertEqual('package', result[1], result)
        # rospack
        (data, type_, url) = locate.get_manifest('rospack', distro)
        self.assertEqual('package', type_)
        data = locate._get_rosinstall_dict('rospack', data, type_)
        self.assertEqual('https://github.com/ros/rospack.git', data.get('git', {}).get('uri', ''))
        # ros_comm
        (data, type_, url) = locate.get_manifest('ros_comm', distro)
        self.assertEqual('metapackage', type_)
        data = locate._get_rosinstall_dict('ros_comm', data, type_)
        self.assertEqual('https://github.com/ros/ros_comm.git', data.get('git', {}).get('uri', ''))



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author: kwc

import sys
import yaml
try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

from catkin_pkg.package import parse_package_string
from rosdistro import get_cached_distribution, get_index, get_index_url

BRANCH_RELEASE = 'release'
BRANCH_DEVEL = 'devel'


class InvalidData(Exception):
    pass

def _get_rosinstall_dict(name, data, type_, branch=None, prefix=None):
    """
    From the dict that was read from the online indexer, create a
    single rosinstall dict.
    """
    # This function takes into account that the way VCS
    # information was colleted by the indexer has varied a lot
    # historically (without documentation or announcement thereof), so
    # it's a mess.
    ri_entry = None
    if branch:
        branch_data = data.get('rosinstalls', None)
        if branch_data:
            ri_entry = branch_data.get(branch, None)
        else:
            sys.stderr.write(
                'Warning: No specific branch data for branch "%s" found, falling back on default checkout\n' % branch)

    # if we were unable to compute the rosinstall info based on a
    # desired branch, use the default info instead
    if ri_entry is None:
        prepared_rosinstall = data.get('rosinstall', None)
        if prepared_rosinstall:
            ri_entry = prepared_rosinstall
        else:
            vcs = get_vcs(name, data, type_)
            vcs_uri = get_vcs_uri(data)
            if not vcs or not vcs_uri:
                raise InvalidData(
                    "Missing VCS control information for %s %s, requires vcs[%s] and vcs_uri[%s]" % (type_, name, vcs, vcs_uri))
            vcs_version = get_vcs_version(data)

            ri_entry = {vcs: {'uri': vcs_uri, 'local-name': name } }

            if vcs_version:
                ri_entry[vcs]['version'] = vcs_version

    if prefix:
        prefix = prefix or ''
        for _, v in ri_entry.items():
            if 'local-name' in v:
                local_name = v['local-name']
                # 3513
                # compute path: we can't use os.path.join because rosinstall paths
                # are always Unix-style.
                paths = [x for x in (prefix, local_name) if x]
                path = '/'.join(paths)
                v['local-name'] = path
    return ri_entry


def get_rosinstall(name, data, type_, branch=None, prefix=None):
    """
    Compute a rosinstall fragment for checkout

    @param name: resource name
    @param data: manifest data for resource
    @param branch: source branch type ('devel' or 'release')
    @param prefix: checkout filepath prefix
    @raise InvalidData
    """
    ri_entry = _get_rosinstall_dict(name, data, type_, branch, prefix)
    return yaml.dump([ri_entry], default_flow_style=False)


def get_vcs_uri_for_branch(data, branch=None):
    """
    @param data: rosdoc manifest data
    @param branch: source branch type ('devel' or 'release')
    """
    ri_entry = None
    if branch:
        branch_data = data.get('rosinstalls', None)
        if branch_data:
            ri_entry = branch_data.get(branch, None)
            vcs_type = list(ri_entry.keys())[0]
            return ri_entry[vcs_type]['uri']
    return data.get('vcs_uri', '')


def get_vcs(name, data, type_):
    """
    @param name: resource name
    @param data: rosdoc manifest data
    @param type_: resource type ('stack' or 'package')
    """
    return data.get('vcs', '')


def get_vcs_version(data):
    return data.get('vcs_version', '')


def get_vcs_uri(data):
    return data.get('vcs_uri', '')


def get_repo(name, data, type_):
    """
    @param name: resource name
    @param data: rosdoc manifest data
    @param type_: resource type ('stack' or 'package')
    """
    return data.get('repo_name', '')


def get_type(data):
    """
    @param data: rosdoc manifest data
    @return 'stack' of 'package'
    """
    return data.get('package_type', 'package')


def get_www(name, data, type_):
    """
    @param name: resource name
    @param data: rosdoc manifest data
    @param type_: resource type ('stack' or 'package')
    """
    return data.get('url', '')


def get_manifest(stackage_name, distro_name=None):
    """
    Get the repository and manifest data.

    @param stackage_name: name of package/stack to get manifest information for.
    get_manifest() gives stacks symbols precedence over package
    symbols.
    @type  stackage_name: str
    @param distro_name: name of ROS distribution
    @type  distro_name: str

    @return: (manifest data, 'package'|'stack'|'repository').
    @rtype: ({str: str}, str, str)
    @raise IOError: if data cannot be loaded
    """
    data = None
    if distro_name is not None:
        data = get_manifest_from_rosdistro(stackage_name, distro_name)
    if data is None:
        sys.stderr.write('Not found via rosdistro - falling back to information provided by rosdoc\n')
        data = get_rosdoc_manifest(stackage_name, distro_name)
    return data


def get_manifest_from_rosdistro(package_name, distro_name):
    """
    Get the rosdistro repository data and package information.

    @param package_name: name of package or repository to get manifest information for.
    It gives package symbols precedence over repository names.
    @type  package_name: str
    @param distro_name: name of ROS distribution
    @type  distro_name: str

    @return: (manifest data, 'package'|'repository').
    @rtype: ({str: str}, str, str)
    @raise IOError: if data cannot be loaded
    """
    data = {}
    type_ = None
    index = get_index(get_index_url())
    try:
        distribution_cache = get_cached_distribution(index, distro_name)
    except RuntimeError as runerr:
        if (runerr.message.startswith("Unknown release")):
            return None
        raise

    if package_name in distribution_cache.release_packages:
        pkg = distribution_cache.release_packages[package_name]
        #print('pkg', pkg.name)
        pkg_xml = distribution_cache.get_release_package_xml(package_name)
        pkg_manifest = parse_package_string(pkg_xml)
        data['description'] = pkg_manifest.description
        website_url = [u.url for u in pkg_manifest.urls if u.type == 'website']
        if website_url:
            data['url'] = website_url[0]
        repo_name = pkg.repository_name
        meta_export = [exp for exp in pkg_manifest.exports if exp.tagname == 'metapackage']
        if meta_export:
            type_ = 'metapackage'
        else:
            type_ = 'package'
    else:
        repo_name = package_name
        type_ = 'repository'
    data['repo_name'] = repo_name
    if repo_name in distribution_cache.repositories:
        repo = distribution_cache.repositories[repo_name].release_repository
        if repo:
            data['packages'] = repo.package_names

    if repo_name in distribution_cache.repositories:
        repo = distribution_cache.repositories[repo_name].source_repository
        if not repo:
            return None
        data['vcs'] = repo.type
        data['vcs_uri'] = repo.url
        data['vcs_version'] = repo.version
    else:
        return None

    return (data, type_, None)


def get_rosdoc_manifest(stackage_name, distro_name=None):
    """
    Get the rosdoc manifest data and type of stackage_name.

    @param stackage_name: name of package/stack to get manifest information for.
    get_manifest() gives stacks symbols precedence over package
    symbols.
    @type  stackage_name: str
    @param distro_name: name of ROS distribution
    @type  distro_name: str

    @return: (manifest data, 'package'|'stack').
    @rtype: ({str: str}, str, str)
    @raise IOError: if data cannot be loaded
    """
    ROSDOC_PREFIX = 'http://ros.org/doc'
    if distro_name is not None:
        prefix = '%s/%s' % (ROSDOC_PREFIX, distro_name)
    else:
        prefix = ROSDOC_PREFIX

    data = None
    url_stack = '%s/api/%s/stack.yaml' % (prefix, stackage_name)
    url_pack = '%s/api/%s/manifest.yaml' % (prefix, stackage_name)
    errors = []
    # ! loop vars used after loop as well
    for type_, url in zip(['stack', 'package'], [url_stack, url_pack]):
        try:
            streamdata = urlopen(url)
            data = yaml.load(streamdata)
            if not data:
                raise InvalidData(
                    'No Information available on %s %s at %s' % (type_,
                                                                 stackage_name,
                                                                 url))
            # with fuerte, stacks also have manifest.yaml, but have a type flag
            realtype = data.get('package_type')
            if realtype:
                type_ = realtype
            break
        except Exception as loope:
            errors.append((url, loope))

    # 1 error is expected when we query package
    if len(errors) > 1:
        error = None
        for (err_url, error) in errors:
            if error is not None:
                sys.stderr.write('error contacting %s:\n%s\n' % (err_url, error))
        raise error
    return (data, type_, url)



# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import unittest
import subprocess
import sys

import rosinstall.helpers
from wstool.config import Config
from wstool.config_yaml import PathSpec

from mock import Mock


class FunctionsTest(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_is_ros_in_setupfile(self):
        try:
            mock_subprocess = Mock()
            mock_process = Mock(name='mockprocess')
            mock_subprocess.Popen.return_value = mock_process
            if sys.version < '3':
                mock_process.communicate.return_value = ('/somewhere/mock_ros_root/foo/..', None)
            else:
                mock_process.communicate.return_value = (b'/somewhere/mock_ros_root/foo/..', None)
            mock_os = Mock()
            mock_path = Mock()
            mock_os.path = mock_path
            mock_os.environ = {}
            mock_path.split = os.path.split
            mock_path.normpath = os.path.normpath
            mock_path.isfile.return_value = True
            rosinstall.helpers.subprocess = mock_subprocess
            rosinstall.helpers.os = mock_os
            result = rosinstall.helpers.get_ros_root_from_setupfile("fooroot/foodir/setup.sh")
            self.assertEqual('/somewhere/mock_ros_root', os.path.normpath(result))
        finally:
            rosinstall.helpers.subprocess = subprocess
            rosinstall.helpers.os = os

    def test_is_path_stack(self):
        self.assertTrue(rosinstall.helpers.is_path_stack(os.path.join("test", "example_dirs", "ros")))
        self.assertTrue(rosinstall.helpers.is_path_stack(os.path.join("test", "example_dirs", "ros_comm")))
        self.assertFalse(rosinstall.helpers.is_path_stack(os.path.join("test", "example_dirs", "roscpp")))

    def test_is_path_ros(self):
        self.assertTrue(rosinstall.helpers.is_path_ros((os.path.join("test", "example_dirs", "ros"))))
        self.assertFalse(rosinstall.helpers.is_path_ros(os.path.join("test", "example_dirs", "ros_comm")))
        self.assertFalse(rosinstall.helpers.is_path_ros((os.path.join("test", "example_dirs", "roscpp"))))

    def test_get_ros_stack_path(self):
        config = Config([PathSpec("foo"),
                         PathSpec(os.path.join("test", "example_dirs", "ros_comm")),
                         PathSpec(os.path.join("test", "example_dirs", "roscpp")),
                         PathSpec("bar")],
                        ".",
                        None)
        self.assertEqual(None, rosinstall.helpers.get_ros_stack_path(config))
        config = Config([PathSpec("foo"),
                         PathSpec(os.path.join("test", "example_dirs", "ros_comm")),
                         PathSpec(os.path.join("test", "example_dirs", "ros")),
                         PathSpec(os.path.join("test", "example_dirs", "roscpp")),
                         PathSpec("bar")],
                        ".",
                        None)
        self.assertEqual(os.path.abspath("test/example_dirs/ros"), rosinstall.helpers.get_ros_stack_path(config))

    def test_get_ros_package_path(self):
        config = Config([],
                        "/test/example_dirs",
                        None)
        self.assertEqual([], rosinstall.helpers.get_ros_package_path(config))
        config = Config([PathSpec("foo")],
                        "/test/example_dirs",
                        None)
        self.assertEqual(['/test/example_dirs/foo'], rosinstall.helpers.get_ros_package_path(config))
        config = Config([PathSpec("foo"),
                         PathSpec(os.path.join("test", "example_dirs", "ros_comm")),
                         PathSpec(os.path.join("test", "example_dirs", "ros")),
                         PathSpec(os.path.join("test", "example_dirs", "roscpp")),
                         PathSpec("bar")],
                        ".",
                        None)
        self.assertEqual(list(map(os.path.abspath, ['bar',
                          'test/example_dirs/roscpp',
                          'test/example_dirs/ros_comm',
                          'foo'])),
                         rosinstall.helpers.get_ros_package_path(config))



# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

import os
import copy
import unittest
import subprocess
import tempfile
import shutil


def _add_to_file(path, content):
    """Util function to append to file to get a modification"""
    with open(path, 'ab') as fhand:
        fhand.write(content.encode('UTF-8'))


def _create_fake_ros_dir(root_path):
    """setup fake ros root within root_path/ros"""
    ros_path = os.path.join(root_path, "ros")
    os.makedirs(ros_path)
    bin_path = os.path.join(ros_path, "bin")
    os.makedirs(bin_path)
    subprocess.check_call(["git", "init"], cwd=ros_path)
    _add_to_file(os.path.join(ros_path, "stack.xml"), '<stack></stack>')
    _add_to_file(os.path.join(ros_path, "setup.sh"), 'export FOO_BAR=`pwd`')
    _add_to_file(os.path.join(bin_path, "rosmake"), '#!/usr/bin/env sh')
    _add_to_file(os.path.join(bin_path, "rospack"), '#!/usr/bin/env sh')
    # even faking rosmake
    subprocess.check_call(["chmod", "u+x", os.path.join(bin_path, "rosmake")])
    subprocess.check_call(["chmod", "u+x", os.path.join(bin_path, "rospack")])
    subprocess.check_call(["git", "add", "*"], cwd=ros_path)
    subprocess.check_call(["git", "commit", "-m", "initial"], cwd=ros_path)


def _create_yaml_file(config_elements, path):
    content = ''
    for elt in list(config_elements):
        content += "- %s:\n" % elt["type"]
        if elt["uri"] is not None:
            content += "    uri: '%s'\n" % elt["uri"]
        content += "    local-name: '%s'\n" % elt["local-name"]
        if elt["version"] is not None:
            content += "    version: '%s'\n" % elt["version"]
    _add_to_file(path, content)


def _create_config_elt_dict(scmtype, localname, uri=None, version=None):
    element = {}
    element["type"]       = scmtype
    element["uri"]        = uri
    element["local-name"] = localname
    element["version"]    = version
    return element


def _create_git_repo(git_path):
    os.makedirs(git_path)
    subprocess.check_call(["git", "init"], cwd=git_path)
    subprocess.check_call(["touch", "gitfixed.txt"], cwd=git_path)
    subprocess.check_call(["git", "add", "*"], cwd=git_path)
    subprocess.check_call(["git", "commit", "-m", "initial"], cwd=git_path)


def _create_tar_file(tar_file):
    parent_path = os.path.dirname(tar_file)
    tar_path = os.path.join(parent_path, 'temptar')
    os.makedirs(tar_path)
    subprocess.check_call(["touch", "tarfixed.txt"], cwd=tar_path)
    subprocess.check_call(["tar", "-czf", os.path.basename(tar_file), 'temptar'], cwd=parent_path)


def _create_hg_repo(hg_path):
    os.makedirs(hg_path)
    subprocess.check_call(["hg", "init"], cwd=hg_path)
    subprocess.check_call(["touch", "hgfixed.txt"], cwd=hg_path)
    subprocess.check_call(["hg", "add", "hgfixed.txt"], cwd=hg_path)
    subprocess.check_call(["hg", "commit", "-m", "initial"], cwd=hg_path)


def _nth_line_split(n, output):
    """returns the last line as list of non-blank tokens"""
    lines = output.splitlines()
    if len(lines) > 0:
        return lines[n].split()
    else:
        return []

# ROSINSTALL_CMD = os.path.join(os.getcwd(), 'scripts/rosinstall')
# ROSWS_CMD = os.path.join(os.getcwd(), 'scripts/rosws')


class AbstractRosinstallCLITest(unittest.TestCase):

    """Base class for cli tests"""
    @classmethod
    def setUpClass(self):
        self.new_environ = copy.copy(os.environ)
        self.new_environ["PYTHONPATH"] = os.path.join(os.getcwd(), "src")
        if "ROS_WORKSPACE" in self.new_environ:
            self.new_environ.pop("ROS_WORKSPACE")


class AbstractRosinstallBaseDirTest(AbstractRosinstallCLITest):
    """test class where each test method get its own fresh tempdir named self.directory"""

    def setUp(self):
        self.directories = {}
        self.directory = tempfile.mkdtemp()
        self.directories["base"] = self.directory
        self.rosinstall_fn = ["rosinstall", "-n"]

    def tearDown(self):
        for d in self.directories:
            shutil.rmtree(self.directories[d])
        self.directories = {}


class AbstractFakeRosBasedTest(AbstractRosinstallBaseDirTest):
    """
    creates some larger infrastructure for testing locally:
    a root folder containing all other folders in self.test_root_path
    a fake ros folder in self.ros_path
    a git repo in self.git_path
    a hg repo in self.hg_path
    a file named self.simple_rosinstall with ros and gitrepo
    a file named self.simple_changed_vcs_rosinstall with ros and hgrepo
    """

    @classmethod
    def setUpClass(self):
        AbstractRosinstallBaseDirTest.setUpClass()
        # create a dir mimicking ros
        self.test_root_path = tempfile.mkdtemp()
        _create_fake_ros_dir(self.test_root_path)
        # create a repo in git
        self.ros_path = os.path.join(self.test_root_path, "ros")
        self.git_path = os.path.join(self.test_root_path, "gitrepo")
        _create_git_repo(self.git_path)
        # create a repo in hg
        self.hg_path = os.path.join(self.test_root_path, "hgrepo")
        _create_hg_repo(self.hg_path)
        # create custom rosinstall files to use as input
        self.simple_rosinstall = os.path.join(self.test_root_path, "simple.rosinstall")
        _create_yaml_file([_create_config_elt_dict("git", "ros", self.ros_path),
                           _create_config_elt_dict("git", "gitrepo", self.git_path)],
                          self.simple_rosinstall)
        self.simple_changed_vcs_rosinstall = os.path.join(self.test_root_path, "simple_changed_vcs.rosinstall")
        _create_yaml_file([_create_config_elt_dict("git", "ros", self.ros_path),
                           _create_config_elt_dict("hg", "hgrepo", self.hg_path)],
                          self.simple_changed_vcs_rosinstall)

    @classmethod
    def tearDownClass(self):
        shutil.rmtree(self.test_root_path)


class AbstractSCMTest(AbstractRosinstallCLITest):
    """Base class for diff tests, setting up a tempdir self.test_root_path for a whole class"""
    @classmethod
    def setUpClass(self):
        """creates a directory 'ros' mimicking to be a ROS root to rosinstall"""
        AbstractRosinstallCLITest.setUpClass()
        self.test_root_path = tempfile.mkdtemp()
        self.directories = {}
        self.directories["root"] = self.test_root_path

        _create_fake_ros_dir(self.test_root_path)
        self.local_path = os.path.join(self.test_root_path, "ws")
        os.makedirs(self.local_path)
        self.curdir = os.getcwd()

    @classmethod
    def tearDownClass(self):
        os.chdir(self.curdir)
        for d in self.directories:
            shutil.rmtree(self.directories[d])

    def assertStatusListEqual(self, listexpect, listactual):
        """helper fun to check scm status output while discarding file ordering differences"""
        lines_expect = listexpect.splitlines()
        lines_actual = listactual.splitlines()
        for line in lines_expect:
            self.assertTrue(line in lines_actual, 'Missing entry %s in output %s' % (line, listactual))
        for line in lines_actual:
            self.assertTrue(line in lines_expect, 'Superflous entry %s in output %s' % (line, listactual))


class UtilTest(unittest.TestCase):
    """test to check the methods run by unit test setups"""

    def test_add_to_file(self):
        self.test_root_path = tempfile.mkdtemp()
        filepath = os.path.join(self.test_root_path, 'foofile')
        self.assertFalse(os.path.exists(filepath))
        _add_to_file(filepath, 'foo')
        self.assertTrue(os.path.exists(filepath))
        with open(filepath, 'r') as f:
            read_data = f.read()
            self.assertEqual(read_data, 'foo')
        _add_to_file(filepath, 'bar')
        with open(filepath, 'r') as f:
            read_data = f.read()
            self.assertEqual(read_data, 'foobar')
        shutil.rmtree(self.test_root_path)

    def test_create_fake_ros(self):
        self.test_root_path = tempfile.mkdtemp()
        rospath = os.path.join(self.test_root_path, 'ros')
        self.assertFalse(os.path.exists(rospath))
        _create_fake_ros_dir(self.test_root_path)
        self.assertTrue(os.path.exists(rospath))
        self.assertTrue(os.path.exists(os.path.join(rospath, "setup.sh")))
        self.assertTrue(os.path.exists(os.path.join(rospath, "stack.xml")))
        self.assertTrue(os.path.exists(os.path.join(rospath, ".git")))
        shutil.rmtree(self.test_root_path)

    def test_create_config_elt_dict(self):
        scmtype = 'foo'
        uri = 'bar'
        localname = 'pip'
        version = 'pop'
        element = _create_config_elt_dict(scmtype, localname, uri, version)
        self.assertEqual(element["type"], scmtype)
        self.assertEqual(element["uri"], uri)
        self.assertEqual(element["local-name"], localname)
        self.assertEqual(element["version"], version)

    def test_create_yaml_file(self):
        self.test_root_path = tempfile.mkdtemp()
        filepath = os.path.join(self.test_root_path, 'foofile')
        config_elements = [
            _create_config_elt_dict("other", "foo"),
            _create_config_elt_dict("git", "foo", "foouri"),
            _create_config_elt_dict("svn", "bar", "baruri", "barversion")]
        _create_yaml_file(config_elements, filepath)
        with open(filepath, 'r') as f:
            read_data = f.read()
            self.assertEqual(read_data, """- other:
    local-name: 'foo'
- git:
    uri: 'foouri'
    local-name: 'foo'
- svn:
    uri: 'baruri'
    local-name: 'bar'
    version: 'barversion'
""")
        shutil.rmtree(self.test_root_path)



class StringIO:
    """
    StringIO.StringIO does not exist in python3
    io.StringIO cannot cope with unicode
    """

    def __init__(self):
        self.stream = ''

    def write(self, data):
        self.stream += data

    def flush(self):
        pass

    def __getattr__(self, attr):
        return getattr(self.stream, attr)

    def getvalue(self):
        return self.stream



#!/usr/bin/env python
# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import subprocess
import shutil
import rosinstall.rosws_cli
from test.scm_test_base import AbstractFakeRosBasedTest, _create_yaml_file, _create_config_elt_dict
from wstool.config_yaml import PathSpec
import wstool.multiproject_cmd
from rosinstall.rosws_cli import RoswsCLI
from rosinstall.helpers import ROSInstallException


class FakeConfig():
    def __init__(self, elts=[], celts=[], basepath=''):
        self.elts = elts
        self.celts = celts
        self.basepath = basepath

    def get_config_elements(self):
        return self.celts

    def get_source(self):
        return self.elts

    def get_base_path(self):
        return self.basepath


class MockConfigElement():
    def __init__(self, local_name='', scmtype=None, path=None, uri=None, spec=None):
        self.scmtype = scmtype
        self.path = path
        self.uri = uri
        self.local_name = local_name
        self.spec = spec

    def get_path(self):
        return self.path

    def get_local_name(self):
        return self.local_name

    def get_path_spec(self):
        return self.spec

    def is_vcs_element(self):
        return True if self.scmtype else False


class RosinstallCommandlineTest(AbstractFakeRosBasedTest):

    def test_require_bootstrap(self):
        config = FakeConfig()
        self.assertFalse(rosinstall.rosinstall_cmd._ros_requires_boostrap(config))
        config = FakeConfig([PathSpec(self.ros_path, path=self.ros_path)])
        self.assertFalse(rosinstall.rosinstall_cmd._ros_requires_boostrap(config))
        config = FakeConfig([PathSpec(self.ros_path, 'git', 'gituri', path=self.ros_path)])
        self.assertTrue(rosinstall.rosinstall_cmd._ros_requires_boostrap(config))


class RosinstallCommandLineGenerationTest(AbstractFakeRosBasedTest):

    def test_cmd_generate_ros_files_simple(self):
        self.local_path = os.path.join(self.test_root_path, "ws")
        os.makedirs(self.local_path)

        config = FakeConfig(celts=[MockConfigElement(path=self.ros_path)], basepath=self.local_path)
        rosinstall.rosinstall_cmd.cmd_generate_ros_files(config, self.local_path, nobuild=True, rosdep_yes=False, catkin=False, catkinpp=None)
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))

    def test_cmd_generate_ros_files_vcs(self):
        self.local_path = os.path.join(self.test_root_path, "ws2")
        os.makedirs(self.local_path)

        config = FakeConfig(celts=[MockConfigElement(path=self.ros_path), MockConfigElement(path='gitrepo', scmtype='git', uri=self.git_path), MockConfigElement(path='hgrepo', scmtype='hg', uri=self.hg_path)], basepath=self.local_path)
        rosinstall.rosinstall_cmd.cmd_generate_ros_files(config, self.local_path, nobuild=True, rosdep_yes=False, catkin=False, catkinpp=None)
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))

    def test_cmd_generate_ros_files_catkin(self):
        self.local_path = os.path.join(self.test_root_path, "ws3")
        os.makedirs(self.local_path)

        config = FakeConfig([PathSpec(self.ros_path), PathSpec('gitrepo', 'git', uri=self.git_path)], self.local_path)
        rosinstall.rosinstall_cmd.cmd_generate_ros_files(config, self.local_path, nobuild=True, rosdep_yes=False, catkin=True, catkinpp=False)
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'CMakeLists.txt')))

    def test_cmd_generate_ros_files_catkinpp(self):
        self.local_path = os.path.join(self.test_root_path, "ws4")
        os.makedirs(self.local_path)

        config = FakeConfig([PathSpec(self.ros_path), PathSpec('gitrepo', 'git', uri=self.git_path)], self.local_path)
        rosinstall.rosinstall_cmd.cmd_generate_ros_files(config, self.local_path, nobuild=True, rosdep_yes=False, catkin=True, catkinpp=True)
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'CMakeLists.txt')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'workspace-config.cmake')))

    def test_cmd_generate_ros_files_build(self):
        self.local_path = os.path.join(self.test_root_path, "ws2b")
        os.makedirs(self.local_path)
        local_rosinstall = os.path.join(self.test_root_path, "local.rosinstall")
        _create_yaml_file([_create_config_elt_dict("git", 'ros_comm', self.git_path),
                           _create_config_elt_dict("git", 'ros', self.ros_path),
                           _create_config_elt_dict("hg", 'hgrepo', self.hg_path)], local_rosinstall)
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([self.local_path, local_rosinstall]))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))

    def test_cmd_init(self):
        self.local_path = os.path.join(self.test_root_path, "ws5")
        os.makedirs(self.local_path)

        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([self.local_path, self.ros_path]))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, '.rosinstall')))
        self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.local_path, 'setup.sh'), shell=True, env=self.new_environ))
        self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.local_path, 'setup.bash'), shell=True, env=self.new_environ, executable='/bin/bash'))

        self.assertEqual(0, cli.cmd_merge(self.local_path, [self.ros_path, "-y"]))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, '.rosinstall')))
        self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.local_path, 'setup.sh'), shell=True, env=self.new_environ))
        self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.local_path, 'setup.bash'), shell=True, env=self.new_environ, executable='/bin/bash'))

    def test_cmd_init_catkin(self):
        self.local_path = os.path.join(self.test_root_path, "ws6")
        os.makedirs(self.local_path)

        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([self.local_path, self.ros_path, "-c"]))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, '.rosinstall')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'CMakeLists.txt')))

    def test_cmd_init_catkin2(self):
        self.local_path = os.path.join(self.test_root_path, "ws7")
        os.makedirs(self.local_path)

        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([self.local_path, self.ros_path, "--catkin"]))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, '.rosinstall')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'CMakeLists.txt')))

    def test_cmd_init_catkinpp(self):
        self.local_path = os.path.join(self.test_root_path, "ws8")
        os.makedirs(self.local_path)

        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([self.local_path, self.ros_path, "--catkin", "--cmake-prefix-path=foo"]))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertFalse(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, '.rosinstall')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'CMakeLists.txt')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'workspace-config.cmake')))

    def test_cmd_init_makedir(self):
        # rosinstall to create dir
        self.local_path = os.path.join(self.test_root_path, "ws9")
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([self.local_path, self.ros_path]))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))

    def test_cmd_init_no_ros(self):
        self.local_path = os.path.join(self.test_root_path, "ws10")

        ros_root_existed = False
        if 'ROS_ROOT' in os.environ:
            ros_root_existed = True
            oldros = os.environ.pop('ROS_ROOT')
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([self.local_path]))

        shutil.rmtree(self.local_path)
        os.environ['ROS_ROOT'] = self.ros_path

        self.assertEqual(0, cli.cmd_init([self.local_path]))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))
        if ros_root_existed:
            os.environ['ROS_ROOT'] = oldros
        else:
            os.environ.pop('ROS_ROOT')

    def test_cmd_init_main(self):
        # rosinstall to create dir
        ros_root_existed = False

        if 'ROS_ROOT' in os.environ:
            ros_root_existed = True
            oldros = os.environ.pop('ROS_ROOT')
        os.environ['ROS_ROOT'] = self.ros_path
        self.local_path = os.path.join(self.test_root_path, "ws11")
        self.assertEqual(0, rosinstall.rosws_cli.rosws_main(['rosws', 'help']))
        self.assertEqual(0, rosinstall.rosws_cli.rosws_main(['rosws', 'init', self.local_path]))
        self.assertEqual(0, rosinstall.rosws_cli.rosws_main(['rosws', 'info', '-t', self.local_path]))
        self.assertEqual(0, rosinstall.rosws_cli.rosws_main(['rosws', 'info', '--pkg-path-only', '-t', self.local_path]))
        self.assertEqual(0, rosinstall.rosws_cli.rosws_main(['rosws', 'info', '--no-pkg-path', '-t', self.local_path]))
        self.assertEqual(0, rosinstall.rosws_cli.rosws_main(['rosws', 'info', '--data-only', '-t', self.local_path]))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.sh')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.bash')))
        self.assertTrue(os.path.exists(os.path.join(self.local_path, 'setup.zsh')))
        if ros_root_existed:
            os.environ['ROS_ROOT'] = oldros
        else:
            os.environ.pop('ROS_ROOT')

    def test_cmd_remove(self):
        # rosinstall to create dir
        self.local_path = os.path.join(self.test_root_path, "ws12")
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([self.local_path, self.ros_path]))
        self.assertEqual(0, cli.cmd_merge(self.local_path, [self.git_path, "-y"]))
        self.assertEqual(0, cli.cmd_merge(self.local_path, [self.hg_path, "-y"]))
        config = wstool.multiproject_cmd.get_config(basepath=self.local_path,
                                                        config_filename='.rosinstall')
        self.assertEqual(len(config.get_config_elements()), 3)
        self.assertEqual(0, cli.cmd_remove(self.local_path, [self.git_path]))
        config = wstool.multiproject_cmd.get_config(basepath=self.local_path,
                                                        config_filename='.rosinstall')
        self.assertEqual(len(config.get_config_elements()), 2)

    def execute_check_result_allshells(self, command, path, cwd=None, expect=''):
        """sources in turn each setup.*sh and runs command"""
        # zsh has issues with SCRIPT_PATH
        for shell in ['sh', 'bash']:
            cmd = ". %s ; %s" % (os.path.join(path, 'setup.%s' % shell), command)
            p = subprocess.Popen(cmd, shell=True, cwd=cwd, stdout=subprocess.PIPE, executable='/bin/%s' % shell)
            output = p.communicate()[0]
            self.assertEqual(expect.encode('UTF-8'), output.strip(), ("'%s' != '%s', cmd: %s, cwd: %s" % (expect, output, cmd, cwd)))
            self.assertEqual(0, p.returncode)

    def test_init_parallel(self):
        self.local_path = os.path.join(self.test_root_path, "ws13a")
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([self.local_path, self.simple_rosinstall, "--parallel=5"]))

        command = "echo $ROS_WORKSPACE"
        self.execute_check_result_allshells(command, self.local_path, expect=self.local_path)
        self.execute_check_result_allshells(command, '.', cwd=self.local_path, expect=self.local_path)
        self.execute_check_result_allshells(command, 'ws13a', cwd=self.test_root_path, expect=self.local_path)

        local_ros_path = os.path.join(self.local_path, "ros")
        local_git_path = os.path.join(self.local_path, "gitrepo")
        package_path = "%s:%s" % (local_git_path, local_ros_path)
        command = "echo $ROS_PACKAGE_PATH"
        self.execute_check_result_allshells(command, self.local_path, expect=package_path)
        self.execute_check_result_allshells(command, '.', cwd=self.local_path, expect=package_path)
        self.execute_check_result_allshells(command, 'ws13a', cwd=self.test_root_path, expect=package_path)

    def test_setup_sh(self):
        self.local_path = os.path.join(self.test_root_path, "ws13")
        cli = RoswsCLI()
        self.assertEqual(0, cli.cmd_init([self.local_path, self.simple_rosinstall]))

        command = "echo $ROS_WORKSPACE"
        self.execute_check_result_allshells(command, self.local_path, expect=self.local_path)
        self.execute_check_result_allshells(command, '.', cwd=self.local_path, expect=self.local_path)
        self.execute_check_result_allshells(command, 'ws13', cwd=self.test_root_path, expect=self.local_path)

        local_ros_path = os.path.join(self.local_path, "ros")
        local_git_path = os.path.join(self.local_path, "gitrepo")
        package_path = "%s:%s" % (local_git_path, local_ros_path)
        command = "echo $ROS_PACKAGE_PATH"
        self.execute_check_result_allshells(command, self.local_path, expect=package_path)
        self.execute_check_result_allshells(command, '.', cwd=self.local_path, expect=package_path)
        self.execute_check_result_allshells(command, 'ws13', cwd=self.test_root_path, expect=package_path)

    def test_setup_sh_relros(self):
        self.local_path = os.path.join(self.test_root_path, "ws14")
        cli = RoswsCLI()
        simple_rel_rosinstall = os.path.join(self.test_root_path, "simple_rel.rosinstall")
        _create_yaml_file([_create_config_elt_dict("git", "ros", "../ros")],
                          simple_rel_rosinstall)
        self.assertEqual(0, cli.cmd_init([self.local_path, simple_rel_rosinstall]))

        command = "echo $ROS_WORKSPACE"
        self.execute_check_result_allshells(command, self.local_path, expect=self.local_path)
        self.execute_check_result_allshells(command, '.', cwd=self.local_path, expect=self.local_path)
        self.execute_check_result_allshells(command, 'ws14', cwd=self.test_root_path, expect=self.local_path)

        package_path = os.path.join(self.local_path, "ros")
        command = "echo $ROS_PACKAGE_PATH"
        self.execute_check_result_allshells(command, self.local_path, expect=package_path)
        self.execute_check_result_allshells(command, '.', cwd=self.local_path, expect=package_path)
        self.execute_check_result_allshells(command, 'ws14', cwd=self.test_root_path, expect=package_path)


    def test_setup_sh_relother(self):
        self.local_path = os.path.join(self.test_root_path, "ws15")
        cli = RoswsCLI()
        simple_rel_rosinstall = os.path.join(self.test_root_path, "simple_rel2.rosinstall")
        _create_yaml_file([_create_config_elt_dict("git", "ros", "../ros"),
                           _create_config_elt_dict("other", "../gitrepo")],
                          simple_rel_rosinstall)
        self.assertEqual(0, cli.cmd_init([self.local_path, simple_rel_rosinstall]))

        command = "echo $ROS_WORKSPACE"
        self.execute_check_result_allshells(command, self.local_path, expect=self.local_path)
        self.execute_check_result_allshells(command, '.', cwd=self.local_path, expect=self.local_path)
        self.execute_check_result_allshells(command, 'ws15', cwd=self.test_root_path, expect=self.local_path)

        local_ros_path = os.path.join(self.local_path, "ros")
        package_path = "%s:%s" % (self.git_path, local_ros_path)
        command = "echo $ROS_PACKAGE_PATH"
        self.execute_check_result_allshells(command, self.local_path, expect=package_path)
        self.execute_check_result_allshells(command, '.', cwd=self.local_path, expect=package_path)
        self.execute_check_result_allshells(command, 'ws15', cwd=self.test_root_path, expect=package_path)



# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import copy
import yaml
import subprocess
import tempfile

import rosinstall
import rosinstall.helpers
from rosinstall.rosinstall_cli import rosinstall_main

from test.scm_test_base import AbstractRosinstallBaseDirTest, AbstractFakeRosBasedTest, _create_yaml_file, _create_config_elt_dict


class RosinstallCommandlineOverlays(AbstractFakeRosBasedTest):
    """test creating parallel rosinstall env with overlayed stacks"""

    def setUp(self):
        """runs rosinstall with generated self.simple_rosinstall to create local rosinstall env
        and creates a directory for a second local rosinstall env"""
        AbstractFakeRosBasedTest.setUp(self)

        # setup a rosinstall env as base for further tests
        cmd = copy.copy(self.rosinstall_fn)
        cmd.extend([self.directory, self.simple_rosinstall])
        self.assertTrue(rosinstall_main(cmd))
        self.new_directory = tempfile.mkdtemp()
        self.directories["new_ros_env"] = self.new_directory

    def test_Rosinstall_rosinstall_file_input(self):
        """uses base rosinstall with ros and git repo"""

        cmd = copy.copy(self.rosinstall_fn)
        cmd.extend([self.new_directory, self.directory])
        self.assertTrue(rosinstall_main(cmd))
        stream = open(os.path.join(self.new_directory, '.rosinstall'), 'r')
        yamlsrc = yaml.load(stream)
        stream.close()
        self.assertEqual(2, len(yamlsrc))
        self.assertEqual('other', list(yamlsrc[0].keys())[0])
        self.assertEqual('other', list(yamlsrc[1].keys())[0])

    def test_Rosinstall_rosinstall_file_input_ros_only(self):
        """uses base ros folder"""
        local_rosinstall = os.path.join(self.test_root_path, "local.rosinstall")
        # invalid recursion itno some other rosinstall folder
        _create_yaml_file([_create_config_elt_dict("other", self.directory)], local_rosinstall)

        cmd = copy.copy(self.rosinstall_fn)
        cmd.extend([self.new_directory, self.ros_path, local_rosinstall])
        self.assertTrue(rosinstall_main(cmd))
        stream = open(os.path.join(self.new_directory, '.rosinstall'), 'r')
        yamlsrc = yaml.load(stream)
        stream.close()
        self.assertEqual(1, len(yamlsrc))
        self.assertEqual('other', list(yamlsrc[0].keys())[0])

    def test_Rosinstall_rosinstall_file_input_add(self):
        """uses base ros folders and adds a stack"""
        local_rosinstall = os.path.join(self.test_root_path, "local2.rosinstall")
        # self.directory points invalidly at a folder containing a .rosinstall pointing to ros and gitrepo
        _create_yaml_file([_create_config_elt_dict("other", self.directory),
                           _create_config_elt_dict("hg", "gitrepo", self.hg_path)],
                          local_rosinstall)

        cmd = copy.copy(self.rosinstall_fn)
        cmd.extend([self.new_directory, self.ros_path, local_rosinstall])
        self.assertTrue(rosinstall_main(cmd))
        stream = open(os.path.join(self.new_directory, '.rosinstall'), 'r')
        yamlsrc = yaml.load(stream)
        stream.close()
        self.assertEqual(2, len(yamlsrc))
        self.assertEqual('other', list(yamlsrc[0].keys())[0])
        self.assertEqual('hg', list(yamlsrc[1].keys())[0])

    def test_Rosinstall_ros_with_folder(self):
        """Use a folder as a remote rosinstall location"""
        cmd = copy.copy(self.rosinstall_fn)
        cmd.extend([self.new_directory, self.directory])
        self.assertTrue(rosinstall_main(cmd))


class RosinstallCommandlineOverlaysWithSetup(AbstractFakeRosBasedTest):
    """test creating parallel rosinstall env with overlayed stacks"""

    def setUp(self):
        """runs rosinstall with generated self.simple_rosinstall to create local rosinstall env
        and creates a second directory self.new_directory for a second local rosinstall env"""
        AbstractFakeRosBasedTest.setUp(self)

        self.simple_fuerte_rosinstall = os.path.join(self.test_root_path, "simple_fuerte.rosinstall")
        _create_yaml_file([_create_config_elt_dict("git", "ros", self.ros_path),
                           _create_config_elt_dict("setup-file", "setup.sh"),
                           _create_config_elt_dict("hg", "hgrepo", self.hg_path)],
                          self.simple_fuerte_rosinstall)

        # setup a rosinstall env as base for further tests
        cmd = copy.copy(self.rosinstall_fn)
        cmd.extend([self.directory, self.simple_fuerte_rosinstall])
        self.assertTrue(rosinstall_main(cmd))

        self.new_directory = tempfile.mkdtemp()
        self.directories["new_ros_env"] = self.new_directory

    def test_Rosinstall_rosinstall_file_input_with_setupfile(self):
        local_rosinstall = os.path.join(self.test_root_path, "local.rosinstall")
        _create_yaml_file([_create_config_elt_dict("other", self.directory),
                           _create_config_elt_dict("hg", "hgrepo", self.hg_path)],
                          local_rosinstall)

        cmd = copy.copy(self.rosinstall_fn)
        cmd.extend([self.new_directory, self.ros_path, local_rosinstall])
        self.assertTrue(rosinstall_main(cmd))
        stream = open(os.path.join(self.new_directory, '.rosinstall'), 'r')
        yamlsrc = yaml.load(stream)
        stream.close()
        self.assertEqual(2, len(yamlsrc), yamlsrc)
        self.assertEqual('other', list(yamlsrc[0].keys())[0])  # ros
        self.assertEqual('hg', list(yamlsrc[1].keys())[0])  # hg_repo

    def test_Rosinstall_ros_with_folder_and_setupfile(self):
        cmd = copy.copy(self.rosinstall_fn)
        cmd.extend([self.new_directory, self.directory])
        self.assertTrue(rosinstall_main(cmd))
        stream = open(os.path.join(self.new_directory, '.rosinstall'), 'r')
        yamlsrc = yaml.load(stream)
        stream.close()
        self.assertEqual(3, len(yamlsrc))
        self.assertEqual('other', list(yamlsrc[0].keys())[0])
        self.assertEqual('setup-file', list(yamlsrc[1].keys())[0])
        self.assertEqual('other', list(yamlsrc[2].keys())[0])


class RosinstallLocalDistro(AbstractRosinstallBaseDirTest):

    def test_prereq(self):
        self.assertTrue(os.path.exists('/bin/bash'))
        self.assertTrue(os.path.exists('/bin/zsh'))

    if os.path.isdir('/opt/ros/cturtle'):
        def test_local_cturtle(self):
            distrodir = '/opt/ros/cturtle'
            cmd = copy.copy(self.rosinstall_fn)
            cmd.extend([self.directory, distrodir])
            self.assertTrue(rosinstall_main(cmd))
            self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.directory, 'setup.sh'), shell=True, env=self.new_environ))
            self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.directory, 'setup.bash'), shell=True, env=self.new_environ, executable='/bin/bash'))

    if os.path.isdir('/opt/ros/diamondback'):
        def test_local_diamondback(self):
            distrodir = '/opt/ros/diamondback'
            cmd = copy.copy(self.rosinstall_fn)
            cmd.extend([self.directory, distrodir])
            self.assertTrue(rosinstall_main(cmd))
            self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.directory, 'setup.sh'), shell=True, env=self.new_environ))
            self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.directory, 'setup.bash'), shell=True, env=self.new_environ, executable='/bin/bash'))
            p = subprocess.Popen("bash -c 'set -e; . %s'" % os.path.join(self.directory, 'setup.bash'),
                                 shell=True,
                                 stdout=subprocess.PIPE,
                                 env=self.new_environ)
            output = p.communicate()
            self.assertEqual(0, p.returncode)
            p = subprocess.Popen("bash -c '. %s; echo $ROS_ROOT'" % os.path.join(self.directory, 'setup.bash'),
                                 shell=True,
                                 stdout=subprocess.PIPE,
                                 env=self.new_environ)
            output = p.communicate()
            self.assertEqual('/opt/ros/diamondback/ros', output[0].decode('UTF-8').rstrip('\n'))
            p = subprocess.Popen("sh -c '. %s; echo $ROS_ROOT'" % os.path.join(self.directory, 'setup.sh'),
                                 shell=True,
                                 stdout=subprocess.PIPE,
                                 env=self.new_environ)
            output = p.communicate()
            self.assertEqual('/opt/ros/diamondback/ros', output[0].decode('UTF-8').rstrip('\n'))

    if os.path.isdir('/opt/ros/electric'):
        def test_local_electric(self):
            distrodir = '/opt/ros/electric'
            cmd = copy.copy(self.rosinstall_fn)
            cmd.extend([self.directory, distrodir])
            self.assertTrue(rosinstall_main(cmd))
            self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.directory, 'setup.sh'), shell=True, env=self.new_environ))
            self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.directory, 'setup.bash'), shell=True, env=self.new_environ, executable='/bin/bash'))
            p = subprocess.Popen("bash -c 'set -e; . %s'" % os.path.join(self.directory, 'setup.bash'),
                                 shell=True,
                                 stdout=subprocess.PIPE,
                                 env=self.new_environ)
            output = p.communicate()
            self.assertEqual(0, p.returncode, (output, p.returncode))
            p = subprocess.Popen("bash -c '. %s; echo $ROS_ROOT'" % os.path.join(self.directory, 'setup.bash'),
                                 shell=True,
                                 stdout=subprocess.PIPE,
                                 env=self.new_environ)
            output = p.communicate()
            self.assertEqual('/opt/ros/electric/ros', output[0].decode('UTF-8').rstrip('\n'))
            p = subprocess.Popen("sh -c '. %s; echo $ROS_ROOT'" % os.path.join(self.directory, 'setup.sh'),
                                 shell=True,
                                 stdout=subprocess.PIPE,
                                 env=self.new_environ)
            output = p.communicate()
            self.assertEqual('/opt/ros/electric/ros', output[0].decode('UTF-8').rstrip('\n'))

    if os.path.isdir('/opt/ros/fuerte'):
        def test_local_fuerte(self):
            distrodir = '/opt/ros/fuerte'
            cmd = copy.copy(self.rosinstall_fn)
            cmd.extend([self.directory, distrodir])
            self.assertTrue(rosinstall_main(cmd))
            self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.directory, 'setup.sh'), shell=True, env=self.new_environ))
            self.assertEqual(0, subprocess.call(". %s" % os.path.join(self.directory, 'setup.bash'), shell=True, env=self.new_environ, executable='/bin/bash'))
            p = subprocess.Popen("bash -c 'set -e; . %s'" % os.path.join(self.directory, 'setup.bash'),
                                 shell=True,
                                 stdout=subprocess.PIPE,
                                 env=self.new_environ)
            output = p.communicate()
            self.assertEqual(0, p.returncode, (output, p.returncode))
            p = subprocess.Popen("bash -c '. %s; echo $ROS_ROOT'" % os.path.join(self.directory, 'setup.bash'),
                                 shell=True,
                                 stdout=subprocess.PIPE,
                                 env=self.new_environ)
            output = p.communicate()
            self.assertEqual('/opt/ros/fuerte/share/ros', output[0].decode('UTF-8').rstrip('\n'))
            p = subprocess.Popen("sh -c '. %s; echo $ROS_ROOT'" % os.path.join(self.directory, 'setup.sh'),
                                 shell=True,
                                 stdout=subprocess.PIPE,
                                 env=self.new_environ)
            output = p.communicate()
            self.assertEqual('/opt/ros/fuerte/share/ros', output[0].decode('UTF-8').rstrip('\n'))

    if os.path.isdir('/opt/ros/fuerte'):
        def test_local_fuerte_catkin(self):
            distrodir = '/opt/ros/fuerte'
            cmd = copy.copy(self.rosinstall_fn)
            cmd.extend([self.directory, distrodir, '--catkin'])
            self.assertTrue(rosinstall_main(cmd))
            self.assertTrue(os.path.exists(os.path.join(self.directory, 'CMakeLists.txt')))



import os
import io
import unittest
import tempfile
import shutil

import rosinstall.helpers
import rosinstall.rosws_stacks_cli
from rosinstall.helpers import ROSInstallException
from wstool.config import Config
from wstool.config_yaml import PathSpec


class RosWsStacksTest(unittest.TestCase):
    def test_get_distro(self):
        self.assertEqual('diamondback', rosinstall.rosws_stacks_cli.rosversion_to_distro_name([1, 4, 0]))
        self.assertEqual('electric', rosinstall.rosws_stacks_cli.rosversion_to_distro_name([1, 6, 0]))
        self.assertEqual('fuerte', rosinstall.rosws_stacks_cli.rosversion_to_distro_name([1, 8, 0]))
        self.assertEqual('groovy', rosinstall.rosws_stacks_cli.rosversion_to_distro_name([1, 10, 0]))
        try:
            self.assertEqual('groovy', rosinstall.rosws_stacks_cli.rosversion_to_distro_name([1]))
            self.fail('expected exception')
        except ROSInstallException:
            pass
        try:
            self.assertEqual('groovy', rosinstall.rosws_stacks_cli.rosversion_to_distro_name(['a', 'b']))
            self.fail('expected exception')
        except ROSInstallException:
            pass

    def test_get_stack_element_in_config(self):
        self.test_root_path = tempfile.mkdtemp()
        self.install_path = os.path.join(self.test_root_path, "install")
        os.makedirs(self.install_path)
        f = io.open(os.path.join(self.install_path, 'stack.xml'), 'a')
        f.write(unicode("hello stack"))
        f.close()
        config = Config([PathSpec("foo"),
                         PathSpec("install"),
                         PathSpec("bar")],
                        self.test_root_path,
                        None)
        self.assertEqual(None, rosinstall.rosws_stacks_cli.get_stack_element_in_config(config, 'foo'))
        self.assertEqual(None, rosinstall.rosws_stacks_cli.get_stack_element_in_config(config, None))
        el = rosinstall.rosws_stacks_cli.get_stack_element_in_config(config, 'install')
        self.assertEqual(self.install_path, el.get_path())
        shutil.rmtree(self.test_root_path)

    def test_roslocate_info(self):
        yaml = rosinstall.rosws_stacks_cli.roslocate_info("ros_comm", "electric", False)
        self.assertEqual([{'svn': {'local-name': 'ros_comm', 'uri': 'https://code.ros.org/svn/ros/stacks/ros_comm/tags/ros_comm-1.6.6'}}], yaml)

    def xtest_get_ros_stack_version(self):
        ver = rosinstall.rosws_stacks_cli.get_ros_stack_version()
        assertTrue(ver[0] > 1)
        assertTrue(ver[1] >= 0)

    def test_get_dependent_stacks(self):
        dep_ons = rosinstall.rosws_stacks_cli.get_dependent_stacks('ros')
        self.assertTrue(len(dep_ons) > 1)



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author: kwc

import rosdistro
from rosdistro.manifest_provider import get_release_tag
from rospkg import distro as rospkg_distro
import yaml
try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

BRANCH_RELEASE = 'release'
BRANCH_DEVEL = 'devel'


class InvalidData(Exception):
    pass


def build_rosinstall(repo_name, uri, vcs_type, version, prefix):
    """
    Build a rosinstall file given some basic information
    """
    rosinstall = []
    repo_name = repo_name if not prefix else '/'.join([prefix, repo_name])

    if version:
        rosinstall.append({vcs_type: {'local-name': repo_name,
                          'uri': uri, 'version': version}})
    else:
        rosinstall.append({vcs_type: {'local-name': repo_name, 'uri': uri}})
    # return yaml.dump(rosinstall, default_flow_style=False)
    return rosinstall


def get_wet_info(wet_distro, name):
    """
    Get information about wet packages or stacks
    """
    repos = wet_distro['repositories']
    for repo in repos:
        info = repos[repo]
        if repo == name or name in info.get('packages', []):
            return (repo, info)
    return None


def get_dry_info(dry_distro, name):
    """
    Get information about dry stacks
    """
    dry_stacks = dry_distro.get_stacks(True)
    if name in dry_stacks:
        stack = dry_stacks[name]
        if stack.vcs_config.type == 'svn':
            return (name,
                    stack.vcs_config.release_tag, stack.vcs_config.type, None)
        else:
            return (name,
                    stack.vcs_config.anon_repo_uri, stack.vcs_config.type,
                    stack.vcs_config.release_tag)
    return None


def get_release_rosinstall(name, wet_distro, dry_distro, prefix):
    """
    Check if the name is in the wet distro
    """
    info = get_wet_info(wet_distro, name)
    if info:
        repo_name, repo_info = info
        if repo_name == name and 'packages' in repo_info:
            rosinstall = []
            pkg_prefix = '/'.join([prefix, repo_name]) if prefix else repo_name
            for pkg in repo_info['packages'].keys():
                rosinstall.extend(build_rosinstall(pkg, repo_info['url'], 'git', '/'.join(
                    ['release', pkg, repo_info['version'].split('-')[0]]), pkg_prefix))
            return rosinstall
        else:
            return build_rosinstall(
                name,
                repo_info['url'],
                'git',
                '/'.join(['release', name, repo_info['version'].split('-')[0]]),
                prefix)

    # Check if the name is in the dry distro
    info = get_dry_info(dry_distro, name)
    if info:
        name, uri, vcs_type, version = info
        return build_rosinstall(name, uri, vcs_type, version, prefix)

    return None


def get_manifest_yaml(name, distro):
    # If we didn't find the name, we need to try to find a stack for it
    url = 'http://ros.org/doc/%s/api/%s/manifest.yaml' % (distro, name)
    try:
        return yaml.load(urlopen(url))
    except:
        raise IOError("Could not load a documentation manifest for %s-%s from ros.org\n\
Have you selected a valid distro? Did you spell everything correctly? Is your package indexed on ros.org?\n\
I'm looking here: %s for a yaml file." % (distro, name, url))


def _get_fuerte_release():
    """
    Please delete me when fuerte is not supported anymore
    See REP137 about rosdistro files
    """
    url = 'https://raw.github.com/ros/rosdistro/master/releases/fuerte.yaml'
    try:
        fuerte_distro = yaml.load(urlopen(url))
    except:
        raise IOError("Could not load the fuerte rosdistro file from github.\n"
                      "Are you sure you've selected a valid distro?\n"
                      "I'm looking for the following file %s" % url)
    return fuerte_distro


def _get_fuerte_rosinstall(name, prefix=None):
    """
    Please delete me when fuerte is not supported anymore
    See REP137 about rosdistro files
    """
    dry_distro = rospkg_distro.load_distro(rospkg_distro.distro_uri('fuerte'))
    wet_distro = _get_fuerte_release()
    # Check to see if the name just exists in one of our rosdistro files
    rosinstall = get_release_rosinstall(name, wet_distro, dry_distro, prefix)
    if rosinstall:
        return rosinstall

    # If we didn't find the name, we need to try to find a stack for it
    doc_yaml = get_manifest_yaml(name, 'fuerte')
    for metapackage in doc_yaml.get('metapackages', []):
        meta_yaml = get_manifest_yaml(metapackage, 'fuerte')
        if meta_yaml['package_type'] == 'stack':
            rosinstall = get_release_rosinstall(
                metapackage, wet_distro, dry_distro, prefix)
            if rosinstall:
                return rosinstall

    return None

def _get_electric_rosinstall(name, prefix=None):
    """
    Please delete me when you don't care at all about electric anymore
    """
    dry_distro = rospkg_distro.load_distro(rospkg_distro.distro_uri('electric'))

    if _is_dry(dry_distro, name):
        return get_dry_rosinstall(dry_distro, name, prefix=prefix)

    # If we didn't find the name, we need to try to find a stack for it
    doc_yaml = get_manifest_yaml(name, 'electric')
    for metapackage in doc_yaml.get('metapackages', []):
        meta_yaml = get_manifest_yaml(metapackage, 'electric')
        if meta_yaml['package_type'] == 'stack':
            if _is_dry(dry_distro, metapackage):
                return get_dry_rosinstall(dry_distro, metapackage, prefix=prefix)

    return None


def _get_rosdistro_release(distro):
    index = rosdistro.get_index(rosdistro.get_index_url())
    return rosdistro.get_distribution_file(index, distro)


def _find_repo(release_file, name):
    for r in release_file.repositories:
        repo = release_file.repositories[r]
        if name in repo.package_names:
            return repo
    return None


def _is_wet(release_file, name):
    return _find_repo(release_file, name) is not None


def _is_dry(dry_distro, name):
    return get_dry_info(dry_distro, name) is not None


def get_wet_rosinstall(release_file, name, prefix=None):
    repo = _find_repo(release_file, name)
    if repo is None:  # wait, what?
        return None
    return build_rosinstall(name, repo.url, 'git', get_release_tag(repo, name), prefix)


def get_dry_rosinstall(dry_distro, name, prefix=None):
    info = get_dry_info(dry_distro, name)
    if info:
        name, uri, vcs_type, version = info
        return build_rosinstall(name, uri, vcs_type, version, prefix)
    return None


def get_release_info(name, distro, prefix=None):
    """
    Steps to check for a released version of the package
    1) Look in the wet distro file for the package/stack name, if it's there, return the repo
    2) Look in the dry distro file for the package/stack name, if it's there, return the repo
    3) Look in the manifest.yaml generated by the documentation indexer to take a best guess at
    what stack a given package belongs to
    4) Look in the distro files again to see if the stack name is there, if it is, return the repo
    """

    # fuerte is different.
    if distro == 'fuerte':
        return _get_fuerte_rosinstall(name, prefix=prefix)

    # electric is ancient.
    if distro == 'electric':
        return _get_electric_rosinstall(name, prefix=prefix)

    wet_distro = _get_rosdistro_release(distro)
    dry_distro = rospkg_distro.load_distro(rospkg_distro.distro_uri(distro))

    # Check to see if the name just exists in one of our rosdistro files
    if _is_wet(wet_distro, name):
        return get_wet_rosinstall(wet_distro, name, prefix=prefix)
    if _is_dry(dry_distro, name):
        return get_dry_rosinstall(dry_distro, name, prefix=prefix)

    # If we didn't find the name, we need to try to find a stack for it
    doc_yaml = get_manifest_yaml(name, distro)
    for metapackage in doc_yaml.get('metapackages', []):
        meta_yaml = get_manifest_yaml(metapackage, distro)
        if meta_yaml['package_type'] == 'stack':
            if _is_dry(dry_distro, metapackage):
                return get_dry_rosinstall(dry_distro, metapackage, prefix=prefix)

    return None


def get_doc_info(name, distro, prefix=None):
    doc_yaml = get_manifest_yaml(name, distro)
    return build_rosinstall(
        doc_yaml['repo_name'], doc_yaml['vcs_uri'], doc_yaml['vcs'],
        doc_yaml.get('vcs_version', ''), prefix)


def get_doc_type(name, distro):
    return get_manifest_yaml(name, distro)['package_type']


def get_doc_www(name, distro):
    return get_manifest_yaml(name, distro)['url']


def get_doc_description(name, distro):
    return get_manifest_yaml(name, distro)['description']
